from .base import ModuleTestBase


class TestSkymem(ModuleTestBase):
    targets = ["blacklanternsecurity.com"]

    async def setup_before_prep(self, module_test):
        module_test.httpx_mock.add_response(
            url="https://www.skymem.info/srch?q=blacklanternsecurity.com",
            text=page_1_body,
        )
        module_test.httpx_mock.add_response(
            url="https://www.skymem.info/domain/5679236812ad5b3f748a413d?p=2",
            text=page_2_body,
        )
        module_test.httpx_mock.add_response(
            url="https://www.skymem.info/domain/5679236812ad5b3f748a413d?p=3",
            text=page_3_body,
        )

    def check(self, module_test, events):
        assert any(e.data == "page1email@blacklanternsecurity.com" for e in events), "Failed to detect first email"
        assert any(e.data == "page2email@blacklanternsecurity.com" for e in events), "Failed to detect second email"
        assert any(e.data == "page3email@blacklanternsecurity.com" for e in events), "Failed to detect third email"


page_1_body = """
<a href="/srch?q=page1email@blacklanternsecurity.com">page1email@blacklanternsecurity.com</a>
<a href="/domain/5679236812ad5b3f748a413d?p=2"><i class="fa fa-arrow-right fa-lg"></i> More emails for <strong>blacklanternsecurity.com </strong> ...</a>
<a href="/domain/5679236812ad5b3f748a413d?p=3"><i class="fa fa-arrow-right fa-lg"></i> More emails for <strong>blacklanternsecurity.com </strong> ...</a>
"""

page_2_body = """
<a href="/srch?q=page2email@blacklanternsecurity.com">page2email@blacklanternsecurity.com</a>
<a href="/domain/5679236812ad5b3f748a413d?p=2"><i class="fa fa-arrow-right fa-lg"></i> More emails for <strong>blacklanternsecurity.com </strong> ...</a>
<a href="/domain/5679236812ad5b3f748a413d?p=3"><i class="fa fa-arrow-right fa-lg"></i> More emails for <strong>blacklanternsecurity.com </strong> ...</a>
"""

page_3_body = """
<a href="/srch?q=page3email@blacklanternsecurity.com">page3email@blacklanternsecurity.com</a>
<a href="/domain/5679236812ad5b3f748a413d?p=2"><i class="fa fa-arrow-right fa-lg"></i> More emails for <strong>blacklanternsecurity.com </strong> ...</a>
<a href="/domain/5679236812ad5b3f748a413d?p=3"><i class="fa fa-arrow-right fa-lg"></i> More emails for <strong>blacklanternsecurity.com </strong> ...</a>
"""



from pathlib import Path
from bbot.modules.base import BaseModule


class apkpure(BaseModule):
    watched_events = ["MOBILE_APP"]
    produced_events = ["FILESYSTEM"]
    flags = ["passive", "safe", "code-enum"]
    meta = {
        "description": "Download android applications from apkpure.com",
        "created_date": "2024-10-11",
        "author": "@domwhewell-sage",
    }
    options = {"output_folder": ""}
    options_desc = {"output_folder": "Folder to download apk's to"}

    async def setup(self):
        output_folder = self.config.get("output_folder")
        if output_folder:
            self.output_dir = Path(output_folder) / "apk_files"
        else:
            self.output_dir = self.scan.home / "apk_files"
        self.helpers.mkdir(self.output_dir)
        return await super().setup()

    async def filter_event(self, event):
        if event.type == "MOBILE_APP":
            if "android" not in event.tags:
                return False, "event is not an android app"
        return True

    async def handle_event(self, event):
        app_id = event.data.get("id", "")
        path = await self.download_apk(app_id)
        if path:
            await self.emit_event(
                {"path": str(path)},
                "FILESYSTEM",
                tags=["apk", "file"],
                parent=event,
                context=f'{{module}} downloaded the apk "{app_id}" to: {path}',
            )

    async def download_apk(self, app_id):
        path = None
        url = f"https://d.apkpure.com/b/XAPK/{app_id}?version=latest"
        self.helpers.mkdir(self.output_dir / app_id)
        file_destination = self.output_dir / app_id / f"{app_id}.xapk"
        result = await self.helpers.download(url, warn=False, filename=file_destination)
        if result:
            self.info(f'Downloaded "{app_id}" from "{url}", saved to {file_destination}')
            path = file_destination
        return path



from .test_module_paramminer_headers import Paramminer_Headers, tempwordlist, helper


class TestParamminer_Cookies(Paramminer_Headers):
    modules_overrides = ["httpx", "paramminer_cookies"]
    config_overrides = {"modules": {"paramminer_cookies": {"wordlist": tempwordlist(["junkcookie", "admincookie"])}}}

    cookies_body = """
    <html>
    <title>the title</title>
    <body>
    <p>Hello null!</p>';
    </body>
    </html>
    """

    cookies_body_match = """
    <html>
    <title>the title</title>
    <body>
    <p>Hello AAAAAAAAAAAAAA!</p>';
    </body>
    </html>
    """

    async def setup_after_prep(self, module_test):
        module_test.scan.modules["paramminer_cookies"].rand_string = lambda *args, **kwargs: "AAAAAAAAAAAAAA"
        module_test.monkeypatch.setattr(
            helper.HttpCompare, "gen_cache_buster", lambda *args, **kwargs: {"AAAAAA": "1"}
        )
        expect_args = dict(headers={"Cookie": "admincookie=AAAAAAAAAAAAAA"})
        respond_args = {"response_data": self.cookies_body_match}
        module_test.set_expect_requests(expect_args=expect_args, respond_args=respond_args)

        respond_args = {"response_data": self.cookies_body}
        module_test.set_expect_requests(respond_args=respond_args)

    def check(self, module_test, events):

        found_reflected_cookie = False
        false_positive_match = False

        for e in events:
            if e.type == "WEB_PARAMETER":
                if "[Paramminer] Cookie: [admincookie] Reasons: [body] Reflection: [True]" in e.data["description"]:
                    found_reflected_cookie = True

                if "junkcookie" in e.data["description"]:
                    false_positive_match = True

        assert found_reflected_cookie, "Failed to find hidden reflected cookie parameter"
        assert not false_positive_match, "Found word which was in wordlist but not a real match"


class TestParamminer_Cookies_noreflection(TestParamminer_Cookies):
    cookies_body_match = """
    <html>
    <title>the title</title>
    <body>
    <p>Hello ADMINISTRATOR!</p>';
    </body>
    </html>
    """

    def check(self, module_test, events):
        assert any(
            e.type == "WEB_PARAMETER"
            and "[Paramminer] Cookie: [admincookie] Reasons: [body] Reflection: [False]" in e.data["description"]
            for e in events
        )



from ..bbot_fixtures import *

from bbot.core.helpers.dns.helpers import extract_targets, service_record, common_srvs


mock_records = {
    "one.one.one.one": {
        "A": ["1.1.1.1", "1.0.0.1"],
        "AAAA": ["2606:4700:4700::1111", "2606:4700:4700::1001"],
        "TXT": [
            '"v=spf1 ip4:103.151.192.0/23 ip4:185.12.80.0/22 ip4:188.172.128.0/20 ip4:192.161.144.0/20 ip4:216.198.0.0/18 ~all"'
        ],
    },
    "1.1.1.1.in-addr.arpa": {"PTR": ["one.one.one.one."]},
}


@pytest.mark.asyncio
async def test_dns_engine(bbot_scanner):
    scan = bbot_scanner()
    await scan.helpers._mock_dns(
        {"one.one.one.one": {"A": ["1.1.1.1"]}, "1.1.1.1.in-addr.arpa": {"PTR": ["one.one.one.one"]}}
    )
    result = await scan.helpers.resolve("one.one.one.one")
    assert "1.1.1.1" in result
    assert not "2606:4700:4700::1111" in result

    results = [_ async for _ in scan.helpers.resolve_batch(("one.one.one.one", "1.1.1.1"))]
    pass_1 = False
    pass_2 = False
    for query, result in results:
        if query == "one.one.one.one" and "1.1.1.1" in result:
            pass_1 = True
        elif query == "1.1.1.1" and "one.one.one.one" in result:
            pass_2 = True
    assert pass_1 and pass_2

    results = [_ async for _ in scan.helpers.resolve_raw_batch((("one.one.one.one", "A"), ("1.1.1.1", "PTR")))]
    pass_1 = False
    pass_2 = False
    for (query, rdtype), (answers, errors) in results:
        results = []
        for answer in answers:
            for t in extract_targets(answer):
                results.append(t[1])
        if query == "one.one.one.one" and "1.1.1.1" in results:
            pass_1 = True
        elif query == "1.1.1.1" and "one.one.one.one" in results:
            pass_2 = True
    assert pass_1 and pass_2

    from bbot.core.helpers.dns.mock import MockResolver

    # ensure dns records are being properly cleaned
    mockresolver = MockResolver({"evilcorp.com": {"MX": ["0 ."]}})
    mx_records = await mockresolver.resolve("evilcorp.com", rdtype="MX")
    results = set()
    for r in mx_records:
        results.update(extract_targets(r))
    assert not results

    await scan._cleanup()


@pytest.mark.asyncio
async def test_dns_resolution(bbot_scanner):
    scan = bbot_scanner("1.1.1.1")

    from bbot.core.helpers.dns.engine import DNSEngine

    dnsengine = DNSEngine(None)
    await dnsengine._mock_dns(mock_records)

    # lowest level functions
    a_responses = await dnsengine._resolve_hostname("one.one.one.one")
    aaaa_responses = await dnsengine._resolve_hostname("one.one.one.one", rdtype="AAAA")
    ip_responses = await dnsengine._resolve_ip("1.1.1.1")
    assert a_responses[0].response.answer[0][0].address in ("1.1.1.1", "1.0.0.1")
    assert aaaa_responses[0].response.answer[0][0].address in ("2606:4700:4700::1111", "2606:4700:4700::1001")
    assert ip_responses[0].response.answer[0][0].target.to_text() in ("one.one.one.one.",)

    # mid level functions
    answers, errors = await dnsengine.resolve_raw("one.one.one.one", type="A")
    responses = []
    for answer in answers:
        responses += list(extract_targets(answer))
    assert ("A", "1.1.1.1") in responses
    assert not ("AAAA", "2606:4700:4700::1111") in responses
    answers, errors = await dnsengine.resolve_raw("one.one.one.one", type="AAAA")
    responses = []
    for answer in answers:
        responses += list(extract_targets(answer))
    assert not ("A", "1.1.1.1") in responses
    assert ("AAAA", "2606:4700:4700::1111") in responses
    answers, errors = await dnsengine.resolve_raw("1.1.1.1")
    responses = []
    for answer in answers:
        responses += list(extract_targets(answer))
    assert ("PTR", "one.one.one.one") in responses

    await dnsengine._shutdown()

    # high level functions
    dnsengine = DNSEngine(None)
    assert "1.1.1.1" in await dnsengine.resolve("one.one.one.one")
    assert "2606:4700:4700::1111" in await dnsengine.resolve("one.one.one.one", type="AAAA")
    assert "one.one.one.one" in await dnsengine.resolve("1.1.1.1")
    for rdtype in ("NS", "SOA", "MX", "TXT"):
        assert len(await dnsengine.resolve("google.com", type=rdtype)) > 0

    # batch resolution
    batch_results = [r async for r in dnsengine.resolve_batch(["1.1.1.1", "one.one.one.one"])]
    assert len(batch_results) == 2
    batch_results = dict(batch_results)
    assert any([x in batch_results["one.one.one.one"] for x in ("1.1.1.1", "1.0.0.1")])
    assert "one.one.one.one" in batch_results["1.1.1.1"]

    # custom batch resolution
    batch_results = [r async for r in dnsengine.resolve_raw_batch([("1.1.1.1", "PTR"), ("one.one.one.one", "A")])]
    batch_results_new = []
    for query, (answers, errors) in batch_results:
        for answer in answers:
            batch_results_new.append((answer.to_text(), answer.rdtype.name))
    assert len(batch_results_new) == 3
    assert any(answer == "1.0.0.1" and rdtype == "A" for answer, rdtype in batch_results_new)
    assert any(answer == "one.one.one.one." and rdtype == "PTR" for answer, rdtype in batch_results_new)

    # dns cache
    dnsengine._dns_cache.clear()
    assert hash(("1.1.1.1", "PTR")) not in dnsengine._dns_cache
    assert hash(("one.one.one.one", "A")) not in dnsengine._dns_cache
    assert hash(("one.one.one.one", "AAAA")) not in dnsengine._dns_cache
    await dnsengine.resolve("1.1.1.1", use_cache=False)
    await dnsengine.resolve("one.one.one.one", use_cache=False)
    assert hash(("1.1.1.1", "PTR")) not in dnsengine._dns_cache
    assert hash(("one.one.one.one", "A")) not in dnsengine._dns_cache
    assert hash(("one.one.one.one", "AAAA")) not in dnsengine._dns_cache

    await dnsengine.resolve("1.1.1.1")
    assert hash(("1.1.1.1", "PTR")) in dnsengine._dns_cache
    await dnsengine.resolve("one.one.one.one", type="A")
    assert hash(("one.one.one.one", "A")) in dnsengine._dns_cache
    assert not hash(("one.one.one.one", "AAAA")) in dnsengine._dns_cache
    dnsengine._dns_cache.clear()
    await dnsengine.resolve("one.one.one.one", type="AAAA")
    assert hash(("one.one.one.one", "AAAA")) in dnsengine._dns_cache
    assert not hash(("one.one.one.one", "A")) in dnsengine._dns_cache

    await dnsengine._shutdown()

    # Ensure events with hosts have resolved_hosts attribute populated
    await scan._prep()
    resolved_hosts_event1 = scan.make_event("one.one.one.one", "DNS_NAME", parent=scan.root_event)
    resolved_hosts_event2 = scan.make_event("http://one.one.one.one/", "URL_UNVERIFIED", parent=scan.root_event)
    dnsresolve = scan.modules["dnsresolve"]
    await dnsresolve.handle_event(resolved_hosts_event1)
    await dnsresolve.handle_event(resolved_hosts_event2)
    assert "1.1.1.1" in resolved_hosts_event2.resolved_hosts
    # URL event should not have dns_children
    assert not resolved_hosts_event2.dns_children
    assert resolved_hosts_event1.resolved_hosts == resolved_hosts_event2.resolved_hosts
    # DNS_NAME event should have dns_children
    assert "1.1.1.1" in resolved_hosts_event1.dns_children["A"]
    assert "A" in resolved_hosts_event1.raw_dns_records
    assert "AAAA" in resolved_hosts_event1.raw_dns_records
    assert "a-record" in resolved_hosts_event1.tags
    assert not "a-record" in resolved_hosts_event2.tags

    scan2 = bbot_scanner("evilcorp.com", config={"dns": {"minimal": False}})
    await scan2.helpers.dns._mock_dns(
        {
            "evilcorp.com": {"TXT": ['"v=spf1 include:cloudprovider.com ~all"']},
            "cloudprovider.com": {"A": ["1.2.3.4"]},
        },
    )
    events = [e async for e in scan2.async_start()]
    assert 1 == len(
        [e for e in events if e.type == "DNS_NAME" and e.data == "cloudprovider.com" and "affiliate" in e.tags]
    )

    await scan._cleanup()
    await scan2._cleanup()


@pytest.mark.asyncio
async def test_wildcards(bbot_scanner):

    scan = bbot_scanner("1.1.1.1")
    helpers = scan.helpers

    from bbot.core.helpers.dns.engine import DNSEngine, all_rdtypes

    dnsengine = DNSEngine(None, debug=True)

    # is_wildcard_domain
    wildcard_domains = await dnsengine.is_wildcard_domain("asdf.github.io", all_rdtypes)
    assert len(dnsengine._wildcard_cache) == len(all_rdtypes) + (len(all_rdtypes) - 2)
    for rdtype in all_rdtypes:
        assert hash(("github.io", rdtype)) in dnsengine._wildcard_cache
        if not rdtype in ("A", "AAAA"):
            assert hash(("asdf.github.io", rdtype)) in dnsengine._wildcard_cache
    assert "github.io" in wildcard_domains
    assert "A" in wildcard_domains["github.io"]
    assert "SRV" not in wildcard_domains["github.io"]
    assert wildcard_domains["github.io"]["A"] and all(helpers.is_ip(r) for r in wildcard_domains["github.io"]["A"][0])
    dnsengine._wildcard_cache.clear()

    # is_wildcard
    for test_domain in ("blacklanternsecurity.github.io", "asdf.asdf.asdf.github.io"):
        wildcard_rdtypes = await dnsengine.is_wildcard(test_domain, all_rdtypes)
        assert "A" in wildcard_rdtypes
        assert "SRV" not in wildcard_rdtypes
        assert wildcard_rdtypes["A"] == (True, "github.io")
        assert wildcard_rdtypes["AAAA"] == (True, "github.io")
        assert len(dnsengine._wildcard_cache) == 2
        for rdtype in ("A", "AAAA"):
            assert hash(("github.io", rdtype)) in dnsengine._wildcard_cache
            assert len(dnsengine._wildcard_cache[hash(("github.io", rdtype))]) == 2
            assert len(dnsengine._wildcard_cache[hash(("github.io", rdtype))][0]) > 0
            assert len(dnsengine._wildcard_cache[hash(("github.io", rdtype))][1]) > 0
        dnsengine._wildcard_cache.clear()

    ### wildcard TXT record ###

    custom_lookup = """
def custom_lookup(query, rdtype):
    if rdtype == "TXT" and query.strip(".").endswith("test.evilcorp.com"):
        return {""}
"""

    mock_data = {
        "evilcorp.com": {"A": ["127.0.0.1"]},
        "test.evilcorp.com": {"A": ["127.0.0.2"]},
        "www.test.evilcorp.com": {"AAAA": ["dead::beef"]},
    }

    # basic sanity checks

    await dnsengine._mock_dns(mock_data, custom_lookup_fn=custom_lookup)

    a_result = await dnsengine.resolve("evilcorp.com")
    assert a_result == {"127.0.0.1"}
    aaaa_result = await dnsengine.resolve("www.test.evilcorp.com", type="AAAA")
    assert aaaa_result == {"dead::beef"}
    txt_result = await dnsengine.resolve("asdf.www.test.evilcorp.com", type="TXT")
    assert txt_result == set()
    txt_result_raw, errors = await dnsengine.resolve_raw("asdf.www.test.evilcorp.com", type="TXT")
    txt_result_raw = list(txt_result_raw)
    assert txt_result_raw

    await dnsengine._shutdown()

    # first, we check with wildcard detection disabled

    scan = bbot_scanner(
        "bbot.fdsa.www.test.evilcorp.com",
        whitelist=["evilcorp.com"],
        config={
            "dns": {"minimal": False, "disable": False, "search_distance": 5, "wildcard_ignore": ["evilcorp.com"]},
            "speculate": True,
        },
    )
    await scan.helpers.dns._mock_dns(mock_data, custom_lookup_fn=custom_lookup)

    events = [e async for e in scan.async_start()]
    assert len(events) == 12
    assert len([e for e in events if e.type == "DNS_NAME"]) == 5
    assert len([e for e in events if e.type == "RAW_DNS_RECORD"]) == 4
    assert sorted([e.data for e in events if e.type == "DNS_NAME"]) == [
        "bbot.fdsa.www.test.evilcorp.com",
        "evilcorp.com",
        "fdsa.www.test.evilcorp.com",
        "test.evilcorp.com",
        "www.test.evilcorp.com",
    ]

    dns_names_by_host = {e.host: e for e in events if e.type == "DNS_NAME"}
    assert dns_names_by_host["evilcorp.com"].tags == {"domain", "private-ip", "in-scope", "a-record"}
    assert dns_names_by_host["evilcorp.com"].resolved_hosts == {"127.0.0.1"}
    assert dns_names_by_host["test.evilcorp.com"].tags == {
        "subdomain",
        "private-ip",
        "in-scope",
        "a-record",
        "txt-record",
    }
    assert dns_names_by_host["test.evilcorp.com"].resolved_hosts == {"127.0.0.2"}
    assert dns_names_by_host["www.test.evilcorp.com"].tags == {"subdomain", "in-scope", "aaaa-record", "txt-record"}
    assert dns_names_by_host["www.test.evilcorp.com"].resolved_hosts == {"dead::beef"}
    assert dns_names_by_host["fdsa.www.test.evilcorp.com"].tags == {"subdomain", "in-scope", "txt-record"}
    assert dns_names_by_host["fdsa.www.test.evilcorp.com"].resolved_hosts == set()
    assert dns_names_by_host["bbot.fdsa.www.test.evilcorp.com"].tags == {
        "target",
        "subdomain",
        "in-scope",
        "txt-record",
    }
    assert dns_names_by_host["bbot.fdsa.www.test.evilcorp.com"].resolved_hosts == set()

    raw_records_by_host = {e.host: e for e in events if e.type == "RAW_DNS_RECORD"}
    assert raw_records_by_host["test.evilcorp.com"].tags == {"subdomain", "in-scope", "txt-record"}
    assert raw_records_by_host["test.evilcorp.com"].resolved_hosts == {"127.0.0.2"}
    assert raw_records_by_host["www.test.evilcorp.com"].tags == {"subdomain", "in-scope", "txt-record"}
    assert raw_records_by_host["www.test.evilcorp.com"].resolved_hosts == {"dead::beef"}
    assert raw_records_by_host["fdsa.www.test.evilcorp.com"].tags == {"subdomain", "in-scope", "txt-record"}
    assert raw_records_by_host["fdsa.www.test.evilcorp.com"].resolved_hosts == set()
    assert raw_records_by_host["bbot.fdsa.www.test.evilcorp.com"].tags == {"subdomain", "in-scope", "txt-record"}
    assert raw_records_by_host["bbot.fdsa.www.test.evilcorp.com"].resolved_hosts == set()

    # then we run it again with wildcard detection enabled

    scan = bbot_scanner(
        "bbot.fdsa.www.test.evilcorp.com",
        whitelist=["evilcorp.com"],
        config={
            "dns": {"minimal": False, "disable": False, "search_distance": 5, "wildcard_ignore": []},
            "speculate": True,
        },
    )
    await scan.helpers.dns._mock_dns(mock_data, custom_lookup_fn=custom_lookup)

    events = [e async for e in scan.async_start()]
    assert len(events) == 12
    assert len([e for e in events if e.type == "DNS_NAME"]) == 5
    assert len([e for e in events if e.type == "RAW_DNS_RECORD"]) == 4
    assert sorted([e.data for e in events if e.type == "DNS_NAME"]) == [
        "_wildcard.test.evilcorp.com",
        "bbot.fdsa.www.test.evilcorp.com",
        "evilcorp.com",
        "test.evilcorp.com",
        "www.test.evilcorp.com",
    ]

    dns_names_by_host = {e.host: e for e in events if e.type == "DNS_NAME"}
    assert dns_names_by_host["evilcorp.com"].tags == {"domain", "private-ip", "in-scope", "a-record"}
    assert dns_names_by_host["evilcorp.com"].resolved_hosts == {"127.0.0.1"}
    assert dns_names_by_host["test.evilcorp.com"].tags == {
        "subdomain",
        "private-ip",
        "in-scope",
        "a-record",
        "txt-record",
    }
    assert dns_names_by_host["test.evilcorp.com"].resolved_hosts == {"127.0.0.2"}
    assert dns_names_by_host["_wildcard.test.evilcorp.com"].tags == {
        "subdomain",
        "in-scope",
        "txt-record",
        "txt-wildcard",
        "wildcard",
    }
    assert dns_names_by_host["_wildcard.test.evilcorp.com"].resolved_hosts == set()
    assert dns_names_by_host["www.test.evilcorp.com"].tags == {
        "subdomain",
        "in-scope",
        "aaaa-record",
        "txt-record",
        "txt-wildcard",
        "wildcard",
    }
    assert dns_names_by_host["www.test.evilcorp.com"].resolved_hosts == {"dead::beef"}
    assert dns_names_by_host["bbot.fdsa.www.test.evilcorp.com"].tags == {
        "target",
        "subdomain",
        "in-scope",
        "txt-record",
        "txt-wildcard",
        "wildcard",
    }
    assert dns_names_by_host["bbot.fdsa.www.test.evilcorp.com"].resolved_hosts == set()

    raw_records_by_host = {e.host: e for e in events if e.type == "RAW_DNS_RECORD"}
    assert raw_records_by_host["test.evilcorp.com"].tags == {"subdomain", "in-scope", "txt-record"}
    assert raw_records_by_host["test.evilcorp.com"].resolved_hosts == {"127.0.0.2"}
    assert raw_records_by_host["www.test.evilcorp.com"].tags == {"subdomain", "in-scope", "txt-record", "txt-wildcard"}
    assert raw_records_by_host["www.test.evilcorp.com"].resolved_hosts == {"dead::beef"}
    assert raw_records_by_host["_wildcard.test.evilcorp.com"].tags == {
        "subdomain",
        "in-scope",
        "txt-record",
        "txt-wildcard",
    }
    assert raw_records_by_host["_wildcard.test.evilcorp.com"].resolved_hosts == set()
    assert raw_records_by_host["bbot.fdsa.www.test.evilcorp.com"].tags == {
        "subdomain",
        "in-scope",
        "txt-record",
        "txt-wildcard",
    }
    assert raw_records_by_host["bbot.fdsa.www.test.evilcorp.com"].resolved_hosts == set()

    ### runaway SRV wildcard ###

    custom_lookup = """
def custom_lookup(query, rdtype):
    if rdtype == "SRV" and query.strip(".").endswith("evilcorp.com"):
        return {f"0 100 389 test.{query}"}
"""

    mock_data = {
        "evilcorp.com": {"A": ["127.0.0.1"]},
        "test.evilcorp.com": {"AAAA": ["dead::beef"]},
    }

    scan = bbot_scanner(
        "evilcorp.com",
        config={
            "dns": {
                "minimal": False,
                "disable": False,
                "search_distance": 5,
                "wildcard_ignore": [],
                "runaway_limit": 3,
            },
        },
    )
    await scan.helpers.dns._mock_dns(mock_data, custom_lookup_fn=custom_lookup)

    events = [e async for e in scan.async_start()]

    assert len(events) == 11
    assert len([e for e in events if e.type == "DNS_NAME"]) == 5
    assert len([e for e in events if e.type == "RAW_DNS_RECORD"]) == 4
    assert sorted([e.data for e in events if e.type == "DNS_NAME"]) == [
        "evilcorp.com",
        "test.evilcorp.com",
        "test.test.evilcorp.com",
        "test.test.test.evilcorp.com",
        "test.test.test.test.evilcorp.com",
    ]

    dns_names_by_host = {e.host: e for e in events if e.type == "DNS_NAME"}
    assert dns_names_by_host["evilcorp.com"].tags == {
        "target",
        "a-record",
        "in-scope",
        "domain",
        "srv-record",
        "private-ip",
    }
    assert dns_names_by_host["test.evilcorp.com"].tags == {
        "in-scope",
        "srv-record",
        "aaaa-record",
        "srv-wildcard-possible",
        "wildcard-possible",
        "subdomain",
    }
    assert dns_names_by_host["test.test.evilcorp.com"].tags == {
        "in-scope",
        "srv-record",
        "srv-wildcard-possible",
        "wildcard-possible",
        "subdomain",
    }
    assert dns_names_by_host["test.test.test.evilcorp.com"].tags == {
        "in-scope",
        "srv-record",
        "srv-wildcard-possible",
        "wildcard-possible",
        "subdomain",
    }
    assert dns_names_by_host["test.test.test.test.evilcorp.com"].tags == {
        "in-scope",
        "srv-record",
        "srv-wildcard-possible",
        "wildcard-possible",
        "subdomain",
        "runaway-dns-3",
    }

    raw_records_by_host = {e.host: e for e in events if e.type == "RAW_DNS_RECORD"}
    assert raw_records_by_host["evilcorp.com"].tags == {"in-scope", "srv-record", "domain"}
    assert raw_records_by_host["test.evilcorp.com"].tags == {
        "in-scope",
        "srv-record",
        "srv-wildcard-possible",
        "subdomain",
    }
    assert raw_records_by_host["test.test.evilcorp.com"].tags == {
        "in-scope",
        "srv-record",
        "srv-wildcard-possible",
        "subdomain",
    }
    assert raw_records_by_host["test.test.test.evilcorp.com"].tags == {
        "in-scope",
        "srv-record",
        "srv-wildcard-possible",
        "subdomain",
    }

    scan = bbot_scanner("1.1.1.1")
    helpers = scan.helpers

    # event resolution
    wildcard_event1 = scan.make_event("wat.asdf.fdsa.github.io", "DNS_NAME", parent=scan.root_event)
    wildcard_event1.scope_distance = 0
    wildcard_event2 = scan.make_event("wats.asd.fdsa.github.io", "DNS_NAME", parent=scan.root_event)
    wildcard_event2.scope_distance = 0
    wildcard_event3 = scan.make_event("github.io", "DNS_NAME", parent=scan.root_event)
    wildcard_event3.scope_distance = 0

    await scan._prep()
    dnsresolve = scan.modules["dnsresolve"]
    await dnsresolve.handle_event(wildcard_event1)
    await dnsresolve.handle_event(wildcard_event2)
    await dnsresolve.handle_event(wildcard_event3)
    assert "wildcard" in wildcard_event1.tags
    assert "a-wildcard" in wildcard_event1.tags
    assert "srv-wildcard" not in wildcard_event1.tags
    assert "wildcard" in wildcard_event2.tags
    assert "a-wildcard" in wildcard_event2.tags
    assert "srv-wildcard" not in wildcard_event2.tags
    assert wildcard_event1.data == "_wildcard.github.io"
    assert wildcard_event2.data == "_wildcard.github.io"
    assert wildcard_event3.data == "github.io"

    # dns resolve distance
    event_distance_0 = scan.make_event(
        "8.8.8.8", module=scan.modules["dnsresolve"]._make_dummy_module("PTR"), parent=scan.root_event
    )
    assert event_distance_0.dns_resolve_distance == 0
    event_distance_1 = scan.make_event(
        "evilcorp.com", module=scan.modules["dnsresolve"]._make_dummy_module("A"), parent=event_distance_0
    )
    assert event_distance_1.dns_resolve_distance == 1
    event_distance_2 = scan.make_event(
        "1.2.3.4", module=scan.modules["dnsresolve"]._make_dummy_module("PTR"), parent=event_distance_1
    )
    assert event_distance_2.dns_resolve_distance == 1
    event_distance_3 = scan.make_event(
        "evilcorp.org", module=scan.modules["dnsresolve"]._make_dummy_module("A"), parent=event_distance_2
    )
    assert event_distance_3.dns_resolve_distance == 2

    await scan._cleanup()

    from bbot.scanner import Scanner

    # test with full scan
    scan2 = Scanner("asdfl.gashdgkjsadgsdf.github.io", whitelist=["github.io"], config={"dns": {"minimal": False}})
    await scan2._prep()
    other_event = scan2.make_event(
        "lkjg.sdfgsg.jgkhajshdsadf.github.io", module=scan2.modules["dnsresolve"], parent=scan2.root_event
    )
    await scan2.ingress_module.queue_event(other_event, {})
    events = [e async for e in scan2.async_start()]
    assert len(events) == 4
    assert 2 == len([e for e in events if e.type == "SCAN"])
    unmodified_wildcard_events = [
        e for e in events if e.type == "DNS_NAME" and e.data == "asdfl.gashdgkjsadgsdf.github.io"
    ]
    assert len(unmodified_wildcard_events) == 1
    assert unmodified_wildcard_events[0].tags.issuperset(
        {
            "a-record",
            "target",
            "aaaa-wildcard",
            "in-scope",
            "subdomain",
            "aaaa-record",
            "wildcard",
            "a-wildcard",
        }
    )
    modified_wildcard_events = [e for e in events if e.type == "DNS_NAME" and e.data == "_wildcard.github.io"]
    assert len(modified_wildcard_events) == 1
    assert modified_wildcard_events[0].tags.issuperset(
        {
            "a-record",
            "aaaa-wildcard",
            "in-scope",
            "subdomain",
            "aaaa-record",
            "wildcard",
            "a-wildcard",
        }
    )
    assert modified_wildcard_events[0].host_original == "lkjg.sdfgsg.jgkhajshdsadf.github.io"

    # test with full scan (wildcard detection disabled for domain)
    scan2 = Scanner(
        "asdfl.gashdgkjsadgsdf.github.io",
        whitelist=["github.io"],
        config={"dns": {"wildcard_ignore": ["github.io"]}},
        exclude_modules=["cloudcheck"],
    )
    await scan2._prep()
    other_event = scan2.make_event(
        "lkjg.sdfgsg.jgkhajshdsadf.github.io", module=scan2.modules["dnsresolve"], parent=scan2.root_event
    )
    await scan2.ingress_module.queue_event(other_event, {})
    events = [e async for e in scan2.async_start()]
    assert len(events) == 4
    assert 2 == len([e for e in events if e.type == "SCAN"])
    unmodified_wildcard_events = [e for e in events if e.type == "DNS_NAME" and "_wildcard" not in e.data]
    assert len(unmodified_wildcard_events) == 2
    assert 1 == len(
        [
            e
            for e in unmodified_wildcard_events
            if e.data == "asdfl.gashdgkjsadgsdf.github.io"
            and e.tags.issuperset(
                {
                    "target",
                    "a-record",
                    "in-scope",
                    "subdomain",
                    "aaaa-record",
                }
            )
        ]
    )
    assert 1 == len(
        [
            e
            for e in unmodified_wildcard_events
            if e.data == "lkjg.sdfgsg.jgkhajshdsadf.github.io"
            and e.tags.issuperset(
                {
                    "a-record",
                    "in-scope",
                    "subdomain",
                    "aaaa-record",
                }
            )
        ]
    )
    modified_wildcard_events = [e for e in events if e.type == "DNS_NAME" and e.data == "_wildcard.github.io"]
    assert len(modified_wildcard_events) == 0


@pytest.mark.asyncio
async def test_dns_raw_records(bbot_scanner):

    from bbot.modules.base import BaseModule

    class DummyModule(BaseModule):
        watched_events = ["*"]

        async def setup(self):
            self.events = []
            return True

        async def handle_event(self, event):
            self.events.append(event)

    # scan without omitted event type
    scan = bbot_scanner("one.one.one.one", "1.1.1.1", config={"dns": {"minimal": False}, "omit_event_types": []})
    await scan.helpers.dns._mock_dns(mock_records)
    dummy_module = DummyModule(scan)
    scan.modules["dummy_module"] = dummy_module
    events = [e async for e in scan.async_start()]
    assert 1 == len([e for e in events if e.type == "RAW_DNS_RECORD"])
    assert 1 == len(
        [
            e
            for e in events
            if e.type == "RAW_DNS_RECORD"
            and e.host == "one.one.one.one"
            and e.data["host"] == "one.one.one.one"
            and e.data["type"] == "TXT"
            and e.data["answer"]
            == '"v=spf1 ip4:103.151.192.0/23 ip4:185.12.80.0/22 ip4:188.172.128.0/20 ip4:192.161.144.0/20 ip4:216.198.0.0/18 ~all"'
            and e.discovery_context == "TXT lookup on one.one.one.one produced RAW_DNS_RECORD"
        ]
    )
    assert 1 == len(
        [
            e
            for e in dummy_module.events
            if e.type == "RAW_DNS_RECORD"
            and e.host == "one.one.one.one"
            and e.data["host"] == "one.one.one.one"
            and e.data["type"] == "TXT"
            and e.data["answer"]
            == '"v=spf1 ip4:103.151.192.0/23 ip4:185.12.80.0/22 ip4:188.172.128.0/20 ip4:192.161.144.0/20 ip4:216.198.0.0/18 ~all"'
            and e.discovery_context == "TXT lookup on one.one.one.one produced RAW_DNS_RECORD"
        ]
    )
    # scan with omitted event type
    scan = bbot_scanner("one.one.one.one", config={"dns": {"minimal": False}, "omit_event_types": ["RAW_DNS_RECORD"]})
    await scan.helpers.dns._mock_dns(mock_records)
    dummy_module = DummyModule(scan)
    scan.modules["dummy_module"] = dummy_module
    events = [e async for e in scan.async_start()]
    # no raw records should be emitted
    assert 0 == len([e for e in events if e.type == "RAW_DNS_RECORD"])
    assert 0 == len([e for e in dummy_module.events if e.type == "RAW_DNS_RECORD"])

    # scan with watching module
    DummyModule.watched_events = ["RAW_DNS_RECORD"]
    scan = bbot_scanner("one.one.one.one", config={"dns": {"minimal": False}, "omit_event_types": ["RAW_DNS_RECORD"]})
    await scan.helpers.dns._mock_dns(mock_records)
    dummy_module = DummyModule(scan)
    scan.modules["dummy_module"] = dummy_module
    events = [e async for e in scan.async_start()]
    # no raw records should be ouptut
    assert 0 == len([e for e in events if e.type == "RAW_DNS_RECORD"])
    # but they should still make it to the module
    assert 1 == len(
        [
            e
            for e in dummy_module.events
            if e.type == "RAW_DNS_RECORD"
            and e.host == "one.one.one.one"
            and e.data["host"] == "one.one.one.one"
            and e.data["type"] == "TXT"
            and e.data["answer"]
            == '"v=spf1 ip4:103.151.192.0/23 ip4:185.12.80.0/22 ip4:188.172.128.0/20 ip4:192.161.144.0/20 ip4:216.198.0.0/18 ~all"'
            and e.discovery_context == "TXT lookup on one.one.one.one produced RAW_DNS_RECORD"
        ]
    )


@pytest.mark.asyncio
async def test_dns_graph_structure(bbot_scanner):
    scan = bbot_scanner("https://evilcorp.com", config={"dns": {"search_distance": 1, "minimal": False}})
    await scan.helpers.dns._mock_dns(
        {
            "evilcorp.com": {
                "CNAME": [
                    "www.evilcorp.com",
                ]
            },
            "www.evilcorp.com": {"CNAME": ["test.evilcorp.com"]},
            "test.evilcorp.com": {"A": ["127.0.0.1"]},
        }
    )
    events = [e async for e in scan.async_start()]
    assert len(events) == 6
    non_scan_events = [e for e in events if e.type != "SCAN"]
    assert sorted([e.type for e in non_scan_events]) == ["DNS_NAME", "DNS_NAME", "DNS_NAME", "URL_UNVERIFIED"]
    events_by_data = {e.data: e for e in non_scan_events}
    assert set(events_by_data) == {"https://evilcorp.com/", "evilcorp.com", "www.evilcorp.com", "test.evilcorp.com"}
    assert events_by_data["test.evilcorp.com"].parent.data == "www.evilcorp.com"
    assert str(events_by_data["test.evilcorp.com"].module) == "CNAME"
    assert events_by_data["www.evilcorp.com"].parent.data == "evilcorp.com"
    assert str(events_by_data["www.evilcorp.com"].module) == "CNAME"
    assert events_by_data["evilcorp.com"].parent.data == "https://evilcorp.com/"
    assert str(events_by_data["evilcorp.com"].module) == "host"


@pytest.mark.asyncio
async def test_dns_helpers(bbot_scanner):
    assert service_record("") == False
    assert service_record("localhost") == False
    assert service_record("www.example.com") == False
    assert service_record("www.example.com", "SRV") == True
    assert service_record("_custom._service.example.com", "SRV") == True
    assert service_record("_custom._service.example.com", "A") == False
    # top 100 most common SRV records
    for srv_record in common_srvs[:100]:
        hostname = f"{srv_record}.example.com"
        assert service_record(hostname) == True

    # make sure system nameservers are excluded from use by DNS brute force
    brute_nameservers = tempwordlist(["1.2.3.4", "8.8.4.4", "4.3.2.1", "8.8.8.8"])
    scan = bbot_scanner(config={"dns": {"brute_nameservers": brute_nameservers}})
    scan.helpers.dns.system_resolvers = ["8.8.8.8", "8.8.4.4"]
    resolver_file = await scan.helpers.dns.brute.resolver_file()
    resolvers = set(scan.helpers.read_file(resolver_file))
    assert resolvers == {"1.2.3.4", "4.3.2.1"}



import multiprocessing
from pathlib import Path
from .base import BaseModule
from badsecrets.base import carve_all_modules


class badsecrets(BaseModule):
    watched_events = ["HTTP_RESPONSE"]
    produced_events = ["FINDING", "VULNERABILITY", "TECHNOLOGY"]
    flags = ["active", "safe", "web-basic"]
    meta = {
        "description": "Library for detecting known or weak secrets across many web frameworks",
        "created_date": "2022-11-19",
        "author": "@liquidsec",
    }
    options = {"custom_secrets": None}
    options_desc = {
        "custom_secrets": "Include custom secrets loaded from a local file",
    }
    deps_pip = ["badsecrets~=0.4.490"]

    async def setup(self):
        self.custom_secrets = None
        custom_secrets = self.config.get("custom_secrets", None)
        if custom_secrets:
            secrets_path = Path(custom_secrets).expanduser()
            if secrets_path.is_file():
                self.custom_secrets = custom_secrets
                self.info(f"Successfully loaded secrets file [{custom_secrets}]")
            else:
                self.warning(f"custom secrets file [{custom_secrets}] is not valid")
                return False, "Custom secrets file not valid"
        return True

    @property
    def _module_threads(self):
        return max(1, multiprocessing.cpu_count() - 1)

    async def handle_event(self, event):
        resp_body = event.data.get("body", None)
        resp_headers = event.data.get("header", None)
        resp_cookies = {}
        if resp_headers:
            resp_cookies_raw = resp_headers.get("set_cookie", None)
            if resp_cookies_raw:
                if "," in resp_cookies_raw:
                    resp_cookies_list = resp_cookies_raw.split(",")
                else:
                    resp_cookies_list = [resp_cookies_raw]
                for c in resp_cookies_list:
                    c2 = c.lstrip(";").strip().split(";")[0].split("=")
                    if len(c2) == 2:
                        resp_cookies[c2[0]] = c2[1]
        if resp_body or resp_cookies:
            try:
                r_list = await self.helpers.run_in_executor_mp(
                    carve_all_modules,
                    body=resp_body,
                    headers=resp_headers,
                    cookies=resp_cookies,
                    url=event.data.get("url", None),
                    custom_resource=self.custom_secrets,
                )
            except Exception as e:
                self.warning(f"Error processing {event}: {e}")
                return
            if r_list:
                for r in r_list:
                    if r["type"] == "SecretFound":
                        data = {
                            "severity": r["description"]["severity"],
                            "description": f"Known Secret Found. Secret Type: [{r['description']['secret']}] Secret: [{r['secret']}] Product Type: [{r['description']['product']}] Product: [{self.helpers.truncate_string(r['product'],2000)}] Detecting Module: [{r['detecting_module']}] Details: [{r['details']}]",
                            "url": event.data["url"],
                            "host": str(event.host),
                        }
                        await self.emit_event(
                            data,
                            "VULNERABILITY",
                            event,
                            context=f'{{module}}\'s "{r["detecting_module"]}" module found known {r["description"]["product"]} secret ({{event.type}}): "{r["secret"]}"',
                        )
                    elif r["type"] == "IdentifyOnly":
                        # There is little value to presenting a non-vulnerable asp.net viewstate, as it is not crackable without a Matrioshka brain. Just emit a technology instead.
                        if r["detecting_module"] == "ASPNET_Viewstate":
                            technology = "microsoft asp.net"
                            await self.emit_event(
                                {"technology": technology, "url": event.data["url"], "host": str(event.host)},
                                "TECHNOLOGY",
                                event,
                                context=f"{{module}} identified {{event.type}}: {technology}",
                            )
                        else:
                            data = {
                                "description": f"Cryptographic Product identified. Product Type: [{r['description']['product']}] Product: [{self.helpers.truncate_string(r['product'],2000)}] Detecting Module: [{r['detecting_module']}]",
                                "url": event.data["url"],
                                "host": str(event.host),
                            }
                            await self.emit_event(
                                data,
                                "FINDING",
                                event,
                                context=f'{{module}} identified cryptographic product ({{event.type}}): "{r["description"]["product"]}"',
                            )



import re
import copy
import logging
import ipaddress
import traceback
from hashlib import sha1
from contextlib import suppress
from radixtarget import RadixTarget

from bbot.errors import *
from bbot.modules.base import BaseModule
from bbot.core.helpers.misc import make_ip_type
from bbot.core.event import make_event, is_event

log = logging.getLogger("bbot.core.target")


class BBOTTarget:
    """
    A convenient abstraction of a scan target that includes whitelisting and blacklisting

    Provides high-level functions like in_scope(), which includes both whitelist and blacklist checks.
    """

    def __init__(self, *targets, whitelist=None, blacklist=None, strict_scope=False, scan=None):
        self.strict_scope = strict_scope
        self.scan = scan
        if len(targets) > 0:
            log.verbose(f"Creating events from {len(targets):,} targets")
        self.seeds = Target(*targets, strict_scope=self.strict_scope, scan=scan)
        if whitelist is None:
            whitelist = set([e.host for e in self.seeds if e.host])
        else:
            log.verbose(f"Creating events from {len(whitelist):,} whitelist entries")
        self.whitelist = Target(*whitelist, strict_scope=self.strict_scope, scan=scan, acl_mode=True)
        if blacklist is None:
            blacklist = []
        if blacklist:
            log.verbose(f"Creating events from {len(blacklist):,} blacklist entries")
        self.blacklist = Target(*blacklist, scan=scan, acl_mode=True)
        self._hash = None

    def add(self, *args, **kwargs):
        self.seeds.add(*args, **kwargs)
        self._hash = None

    def get(self, host):
        return self.seeds.get(host)

    def get_host(self, host):
        return self.seeds.get(host)

    def __iter__(self):
        return iter(self.seeds)

    def __len__(self):
        return len(self.seeds)

    def __contains__(self, other):
        if isinstance(other, self.__class__):
            other = other.seeds
        return other in self.seeds

    def __bool__(self):
        return bool(self.seeds)

    def __eq__(self, other):
        return self.hash == other.hash

    @property
    def hash(self):
        """
        A sha1 hash representing a BBOT target and all three of its components (seeds, whitelist, blacklist)

        This can be used to compare targets.

        Examples:
            >>> target1 = BBOTTarget("evilcorp.com", blacklist=["prod.evilcorp.com"], whitelist=["test.evilcorp.com"])
            >>> target2 = BBOTTarget("evilcorp.com", blacklist=["prod.evilcorp.com"], whitelist=["test.evilcorp.com"])
            >>> target3 = BBOTTarget("evilcorp.com", blacklist=["prod.evilcorp.com"])
            >>> target1 == target2
            True
            >>> target1 == target3
            False
        """
        if self._hash is None:
            # Create a new SHA-1 hash object
            sha1_hash = sha1()
            # Update the SHA-1 object with the hash values of each object
            for target_hash in [t.hash for t in (self.seeds, self.whitelist, self.blacklist)]:
                # Convert the hash value to bytes and update the SHA-1 object
                sha1_hash.update(target_hash)
            self._hash = sha1_hash.digest()
        return self._hash

    @property
    def scope_hash(self):
        """
        A sha1 hash representing only the whitelist and blacklist

        This is used to record the scope of a scan.
        """
        # Create a new SHA-1 hash object
        sha1_hash = sha1()
        # Update the SHA-1 object with the hash values of each object
        for target_hash in [t.hash for t in (self.whitelist, self.blacklist)]:
            # Convert the hash value to bytes and update the SHA-1 object
            sha1_hash.update(target_hash)
        return sha1_hash.digest()

    @property
    def json(self):
        return {
            "seeds": sorted([e.data for e in self.seeds]),
            "whitelist": sorted([e.data for e in self.whitelist]),
            "blacklist": sorted([e.data for e in self.blacklist]),
            "strict_scope": self.strict_scope,
            "hash": self.hash.hex(),
            "seed_hash": self.seeds.hash.hex(),
            "whitelist_hash": self.whitelist.hash.hex(),
            "blacklist_hash": self.blacklist.hash.hex(),
            "scope_hash": self.scope_hash.hex(),
        }

    def copy(self):
        self_copy = copy.copy(self)
        self_copy.seeds = self.seeds.copy()
        self_copy.whitelist = self.whitelist.copy()
        self_copy.blacklist = self.blacklist.copy()
        return self_copy

    @property
    def events(self):
        return self.seeds.events

    def in_scope(self, host):
        """
        Check whether a hostname, url, IP, etc. is in scope.
        Accepts either events or string data.

        Checks whitelist and blacklist.
        If `host` is an event and its scope distance is zero, it will automatically be considered in-scope.

        Examples:
            Check if a URL is in scope:
            >>> preset.in_scope("http://www.evilcorp.com")
            True
        """
        try:
            e = make_event(host, dummy=True)
        except ValidationError:
            return False
        in_scope = e.scope_distance == 0 or self.whitelisted(e)
        return in_scope and not self.blacklisted(e)

    def blacklisted(self, host):
        """
        Check whether a hostname, url, IP, etc. is blacklisted.

        Note that `host` can be a hostname, IP address, CIDR, email address, or any BBOT `Event` with the `host` attribute.

        Args:
            host (str or IPAddress or Event): The host to check against the blacklist

        Examples:
            Check if a URL's host is blacklisted:
            >>> preset.blacklisted("http://www.evilcorp.com")
            True
        """
        e = make_event(host, dummy=True)
        return e in self.blacklist

    def whitelisted(self, host):
        """
        Check whether a hostname, url, IP, etc. is whitelisted.

        Note that `host` can be a hostname, IP address, CIDR, email address, or any BBOT `Event` with the `host` attribute.

        Args:
            host (str or IPAddress or Event): The host to check against the whitelist

        Examples:
            Check if a URL's host is whitelisted:
            >>> preset.whitelisted("http://www.evilcorp.com")
            True
        """
        e = make_event(host, dummy=True)
        whitelist = self.whitelist
        if whitelist is None:
            whitelist = self.seeds
        return e in whitelist

    @property
    def radix_only(self):
        """
        A slimmer, serializable version of the target designed for simple scope checks

        This version doesn't have the events, only their hosts.
        """
        return self.__class__(
            *[e.host for e in self.seeds if e.host],
            whitelist=None if self.whitelist is None else [e for e in self.whitelist],
            blacklist=[e for e in self.blacklist],
            strict_scope=self.strict_scope,
        )


class Target:
    """
    A class representing a target. Can contain an unlimited number of hosts, IP or IP ranges, URLs, etc.

    Attributes:
        strict_scope (bool): Flag indicating whether to consider child domains in-scope.
            If set to True, only the exact hosts specified and not their children are considered part of the target.

        _radix (RadixTree): Radix tree for quick IP/DNS lookups.
        _events (set): Flat set of contained events.

    Examples:
        Basic usage
        >>> target = Target(scan, "evilcorp.com", "1.2.3.0/24")
        >>> len(target)
        257
        >>> list(t.events)
        [
            DNS_NAME("evilcorp.com", module=TARGET, tags={'domain', 'distance-1', 'target'}),
            IP_RANGE("1.2.3.0/24", module=TARGET, tags={'ipv4', 'distance-1', 'target'})
        ]
        >>> "www.evilcorp.com" in target
        True
        >>> "1.2.3.4" in target
        True
        >>> "4.3.2.1" in target
        False
        >>> "https://admin.evilcorp.com" in target
        True
        >>> "bob@evilcorp.com" in target
        True

        Event correlation
        >>> target.get("www.evilcorp.com")
        DNS_NAME("evilcorp.com", module=TARGET, tags={'domain', 'distance-1', 'target'})
        >>> target.get("1.2.3.4")
        IP_RANGE("1.2.3.0/24", module=TARGET, tags={'ipv4', 'distance-1', 'target'})

        Target comparison
        >>> target2 = Targets(scan, "www.evilcorp.com")
        >>> target2 == target
        False
        >>> target2 in target
        True
        >>> target in target2
        False

    Notes:
        - Targets are only precise down to the individual host. Ports and protocols are not considered in scope calculations.
        - If you specify "https://evilcorp.com:8443" as a target, all of evilcorp.com (including subdomains and other ports and protocols) will be considered part of the target
        - If you do not want to include child subdomains, use `strict_scope=True`
    """

    def __init__(self, *targets, strict_scope=False, scan=None, acl_mode=False):
        """
        Initialize a Target object.

        Args:
            *targets: One or more targets (e.g., domain names, IP ranges) to be included in this Target.
            strict_scope (bool): Whether to consider subdomains of target domains in-scope
            scan (Scan): Reference to the Scan object that instantiated the Target.
            acl_mode (bool): Stricter deduplication for more efficient checks

        Notes:
            - If you are instantiating a target from within a BBOT module, use `self.helpers.make_target()` instead. (this removes the need to pass in a scan object.)
            - The strict_scope flag can be set to restrict scope calculation to only exactly-matching hosts and not their child subdomains.
            - Each target is processed and stored as an `Event` in the '_events' dictionary.
        """
        self.scan = scan
        self.strict_scope = strict_scope
        self.acl_mode = acl_mode
        self.special_event_types = {
            "ORG_STUB": re.compile(r"^(?:ORG|ORG_STUB):(.*)", re.IGNORECASE),
            "USERNAME": re.compile(r"^(?:USER|USERNAME):(.*)", re.IGNORECASE),
        }
        self._events = set()
        self._radix = RadixTarget()

        for target_event in self._make_events(targets):
            self._add_event(target_event)

        self._hash = None

    def add(self, t, event_type=None):
        """
        Add a target or merge events from another Target object into this Target.

        Args:
            t: The target to be added. It can be either a string, an event object, or another Target object.

        Attributes Modified:
            _events (dict): The dictionary is updated to include the new target's events.

        Examples:
            >>> target.add('example.com')

        Notes:
            - If `t` is of the same class as this Target, all its events are merged.
            - If `t` is an event, it is directly added to `_events`.
        """
        if not isinstance(t, (list, tuple, set)):
            t = [t]
        for single_target in t:
            if isinstance(single_target, self.__class__):
                for event in single_target.events:
                    self._add_event(event)
            else:
                if is_event(single_target):
                    event = single_target
                else:
                    try:
                        event = make_event(
                            single_target, event_type=event_type, dummy=True, tags=["target"], scan=self.scan
                        )
                    except ValidationError as e:
                        # allow commented lines
                        if not str(t).startswith("#"):
                            log.trace(traceback.format_exc())
                            raise ValidationError(f'Could not add target "{t}": {e}')
                self._add_event(event)

    @property
    def events(self):
        """
        Returns all events in the target.

        Yields:
            Event object: One of the Event objects stored in the `_events` dictionary.

        Examples:
            >>> target = Target(scan, "example.com")
            >>> for event in target.events:
            ...     print(event)

        Notes:
            - This property is read-only.
        """
        return self._events

    @property
    def hosts(self):
        return [e.host for e in self.events]

    def copy(self):
        """
        Creates and returns a copy of the Target object, including a shallow copy of the `_events` and `_radix` attributes.

        Returns:
            Target: A new Target object with the sameattributes as the original.
                    A shallow copy of the `_events` dictionary is made.

        Examples:
            >>> original_target = Target(scan, "example.com")
            >>> copied_target = original_target.copy()
            >>> copied_target is original_target
            False
            >>> copied_target == original_target
            True
            >>> copied_target in original_target
            True
            >>> original_target in copied_target
            True

        Notes:
            - The `scan` object reference is kept intact in the copied Target object.
        """
        self_copy = self.__class__()
        self_copy._events = set(self._events)
        self_copy._radix = copy.copy(self._radix)
        return self_copy

    def get(self, host, single=True):
        """
        Gets the event associated with the specified host from the target's radix tree.

        Args:
            host (Event, Target, or str): The hostname, IP, URL, or event to look for.
            single (bool): Whether to return a single event. If False, return all events matching the host

        Returns:
            Event or None: Returns the Event object associated with the given host if it exists, otherwise returns None.

        Examples:
            >>> target = Target(scan, "evilcorp.com", "1.2.3.0/24")
            >>> target.get("www.evilcorp.com")
            DNS_NAME("evilcorp.com", module=TARGET, tags={'domain', 'distance-1', 'target'})
            >>> target.get("1.2.3.4")
            IP_RANGE("1.2.3.0/24", module=TARGET, tags={'ipv4', 'distance-1', 'target'})

        Notes:
            - The method returns the first event that matches the given host.
            - If `strict_scope` is False, it will also consider parent domains and IP ranges.
        """
        try:
            event = make_event(host, dummy=True)
        except ValidationError:
            return
        if event.host:
            return self.get_host(event.host, single=single)

    def get_host(self, host, single=True):
        """
        A more efficient version of .get() that only accepts hostnames and IP addresses
        """
        host = make_ip_type(host)
        with suppress(KeyError, StopIteration):
            result = self._radix.search(host)
            if result is not None:
                ret = set()
                for event in result:
                    # if the result is a dns name and strict scope is enabled
                    if isinstance(event.host, str) and self.strict_scope:
                        # if the result doesn't exactly equal the host, abort
                        if event.host != host:
                            return
                    if single:
                        return event
                    else:
                        ret.add(event)
                if ret and not single:
                    return ret

    def _sort_events(self, events):
        return sorted(events, key=lambda x: x._host_size)

    def _make_events(self, targets):
        events = []
        for target in targets:
            event_type = None
            for eventtype, regex in self.special_event_types.items():
                if isinstance(target, str):
                    match = regex.match(target)
                    if match:
                        target = match.groups()[0]
                        event_type = eventtype
                        break
            events.append(make_event(target, event_type=event_type, dummy=True, scan=self.scan))
        return self._sort_events(events)

    def _add_event(self, event):
        skip = False
        if event.host:
            radix_data = self._radix.search(event.host)
            if self.acl_mode:
                # skip if the hostname/IP/subnet (or its parent) has already been added
                if radix_data is not None and not self.strict_scope:
                    skip = True
                else:
                    event_type = "IP_RANGE" if event.type == "IP_RANGE" else "DNS_NAME"
                    event = make_event(event.host, event_type=event_type, dummy=True, scan=self.scan)
            if not skip:
                # if strict scope is enabled and it's not an exact host match, we add a whole new entry
                if radix_data is None or (self.strict_scope and event.host not in radix_data):
                    radix_data = {event}
                    self._radix.insert(event.host, radix_data)
                # otherwise, we add the event to the set
                else:
                    radix_data.add(event)
                # clear hash
                self._hash = None
        elif self.acl_mode and not self.strict_scope:
            # skip if we're in ACL mode and there's no host
            skip = True
        if not skip:
            self._events.add(event)

    def _contains(self, other):
        if self.get(other) is not None:
            return True
        return False

    def __str__(self):
        return ",".join([str(e.data) for e in self.events][:5])

    def __iter__(self):
        yield from self.events

    def __contains__(self, other):
        # if "other" is a Target
        if isinstance(other, self.__class__):
            contained_in_self = [self._contains(e) for e in other.events]
            return all(contained_in_self)
        else:
            return self._contains(other)

    def __bool__(self):
        return bool(self._events)

    def __eq__(self, other):
        return self.hash == other.hash

    @property
    def hash(self):
        if self._hash is None:
            # Create a new SHA-1 hash object
            sha1_hash = sha1()
            # Update the SHA-1 object with the hash values of each object
            for event_type, event_hash in sorted([(e.type.encode(), e.data_hash) for e in self.events]):
                sha1_hash.update(event_type)
                sha1_hash.update(event_hash)
            if self.strict_scope:
                sha1_hash.update(b"\x00")
            self._hash = sha1_hash.digest()
        return self._hash

    def __len__(self):
        """
        Calculates and returns the total number of hosts within this target, not counting duplicate events.

        Returns:
            int: The total number of unique hosts present within the target's `_events`.

        Examples:
            >>> target = Target(scan, "evilcorp.com", "1.2.3.0/24")
            >>> len(target)
            257

        Notes:
            - If a host is represented as an IP network, all individual IP addresses in that network are counted.
            - For other types of hosts, each unique event is counted as one.
        """
        num_hosts = 0
        for event in self._events:
            if isinstance(event.host, (ipaddress.IPv4Network, ipaddress.IPv6Network)):
                num_hosts += event.host.num_addresses
            else:
                num_hosts += 1
        return num_hosts


class TargetDummyModule(BaseModule):
    _type = "TARGET"
    name = "TARGET"

    def __init__(self, scan):
        self.scan = scan



from baddns.base import get_all_modules
from baddns.lib.loader import load_signatures
from .base import BaseModule

import logging


class baddns_direct(BaseModule):
    watched_events = ["URL", "STORAGE_BUCKET"]
    produced_events = ["FINDING", "VULNERABILITY"]
    flags = ["active", "safe", "subdomain-enum", "baddns", "cloud-enum"]
    meta = {
        "description": "Check for unusual subdomain / service takeover edge cases that require direct detection",
        "created_date": "2024-01-29",
        "author": "@liquidsec",
    }
    options = {"custom_nameservers": []}
    options_desc = {
        "custom_nameservers": "Force BadDNS to use a list of custom nameservers",
    }
    module_threads = 8
    deps_pip = ["baddns~=1.1.864"]

    scope_distance_modifier = 1

    async def setup(self):
        self.preset.core.logger.include_logger(logging.getLogger("baddns"))
        self.custom_nameservers = self.config.get("custom_nameservers", []) or None
        if self.custom_nameservers:
            self.custom_nameservers = self.helpers.chain_lists(self.custom_nameservers)
        self.only_high_confidence = self.config.get("only_high_confidence", False)
        self.signatures = load_signatures()
        return True

    def select_modules(self):
        selected_modules = []
        for m in get_all_modules():
            if m.name in ["CNAME"]:
                selected_modules.append(m)
        return selected_modules

    async def handle_event(self, event):
        CNAME_direct_module = self.select_modules()[0]
        kwargs = {
            "http_client_class": self.scan.helpers.web.AsyncClient,
            "dns_client": self.scan.helpers.dns.resolver,
            "custom_nameservers": self.custom_nameservers,
            "signatures": self.signatures,
            "direct_mode": True,
        }

        CNAME_direct_instance = CNAME_direct_module(event.host, **kwargs)
        if await CNAME_direct_instance.dispatch():

            results = CNAME_direct_instance.analyze()
            if results and len(results) > 0:
                for r in results:
                    r_dict = r.to_dict()

                    data = {
                        "description": f"Possible [{r_dict['signature']}] via direct BadDNS analysis. Indicator: [{r_dict['indicator']}] Trigger: [{r_dict['trigger']}] baddns Module: [{r_dict['module']}]",
                        "host": str(event.host),
                    }

                    await self.emit_event(
                        data,
                        "FINDING",
                        event,
                        tags=[f"baddns-{CNAME_direct_module.name.lower()}"],
                        context=f'{{module}}\'s "{r_dict["module"]}" module found {{event.type}}: {r_dict["description"]}',
                    )
        await CNAME_direct_instance.cleanup()

    async def filter_event(self, event):
        if event.type == "STORAGE_BUCKET":
            if str(event.module).startswith("bucket_"):
                return False
            self.debug(f"Processing STORAGE_BUCKET for {event.host}")
        if event.type == "URL":
            if event.scope_distance > 0:
                self.debug(
                    f"Rejecting {event.host} due to not being in scope (scope distance: {str(event.scope_distance)})"
                )
                return False
            if "cdn-cloudflare" not in event.tags:
                self.debug(f"Rejecting {event.host} due to not being behind CloudFlare")
                return False
            if "status-200" in event.tags or "status-301" in event.tags:
                self.debug(f"Rejecting {event.host} due to lack of non-standard status code")
                return False

            self.debug(f"Passed all checks and is processing {event.host}")
        return True



from ..bbot_fixtures import *  # noqa F401

from bbot.scanner import Scanner, Preset


# FUTURE TODO:
# Consider testing possible edge cases:
#  make sure custom module load directory works with cli arg module/flag/config syntax validation
#   what if you specify -c modules.custommodule.option?
#    the validation needs to not happen until after your custom preset preset has been loaded
#   what if you specify flags in one preset, but another preset (loaded later) has more custom modules that match that flag?
#    how do we make sure those other modules get loaded too?
#   what if you specify a flag that's only on custom modules? Will it be rejected as invalid?


def test_preset_descriptions():
    # ensure very preset has a description
    preset = Preset()
    for yaml_file, (loaded_preset, category, preset_path, original_filename) in preset.all_presets.items():
        assert (
            loaded_preset.description
        ), f'Preset "{loaded_preset.name}" at {original_filename} does not have a description.'


def test_core():
    from bbot.core import CORE

    import omegaconf

    assert "testasdf" not in CORE.default_config
    assert "testasdf" not in CORE.custom_config
    assert "testasdf" not in CORE.config

    core_copy = CORE.copy()
    # make sure our default config is read-only
    with pytest.raises(omegaconf.errors.ReadonlyConfigError):
        core_copy.default_config["testasdf"] = "test"
    # same for merged config
    with pytest.raises(omegaconf.errors.ReadonlyConfigError):
        core_copy.config["testasdf"] = "test"

    assert "testasdf" not in core_copy.default_config
    assert "testasdf" not in core_copy.custom_config
    assert "testasdf" not in core_copy.config

    core_copy.custom_config["testasdf"] = "test"
    assert "testasdf" not in core_copy.default_config
    assert "testasdf" in core_copy.custom_config
    assert "testasdf" in core_copy.config

    # test config merging
    config_to_merge = omegaconf.OmegaConf.create({"test123": {"test321": [3, 2, 1], "test456": [4, 5, 6]}})
    core_copy.merge_custom(config_to_merge)
    assert "test123" not in core_copy.default_config
    assert "test123" in core_copy.custom_config
    assert "test123" in core_copy.config
    assert "test321" in core_copy.custom_config["test123"]
    assert "test321" in core_copy.config["test123"]

    # test deletion
    del core_copy.custom_config.test123.test321
    assert "test123" in core_copy.custom_config
    assert "test123" in core_copy.config
    assert "test321" not in core_copy.custom_config["test123"]
    assert "test321" not in core_copy.config["test123"]
    assert "test456" in core_copy.custom_config["test123"]
    assert "test456" in core_copy.config["test123"]


def test_preset_yaml(clean_default_config):

    import yaml

    preset1 = Preset(
        "evilcorp.com",
        "www.evilcorp.ce",
        whitelist=["evilcorp.ce"],
        blacklist=["test.www.evilcorp.ce"],
        modules=["sslcert"],
        output_modules=["json"],
        exclude_modules=["ipneighbor"],
        flags=["subdomain-enum"],
        require_flags=["safe"],
        exclude_flags=["slow"],
        verbose=False,
        debug=False,
        silent=True,
        config={"preset_test_asdf": 1},
        strict_scope=False,
    )
    preset1 = preset1.bake()
    assert "evilcorp.com" in preset1.target
    assert "evilcorp.ce" in preset1.whitelist
    assert "test.www.evilcorp.ce" in preset1.blacklist
    assert "sslcert" in preset1.scan_modules
    assert preset1.whitelisted("evilcorp.ce")
    assert preset1.whitelisted("www.evilcorp.ce")
    assert not preset1.whitelisted("evilcorp.com")
    assert preset1.blacklisted("test.www.evilcorp.ce")
    assert preset1.blacklisted("asdf.test.www.evilcorp.ce")
    assert not preset1.blacklisted("www.evilcorp.ce")

    # test yaml save/load
    yaml1 = preset1.to_yaml(sort_keys=True)
    preset2 = Preset.from_yaml_string(yaml1)
    yaml2 = preset2.to_yaml(sort_keys=True)
    assert yaml1 == yaml2

    yaml_string_1 = """
flags:
  - subdomain-enum

exclude_flags:
  - aggressive
  - slow

require_flags:
  - passive
  - safe

exclude_modules:
  - certspotter
  - rapiddns

modules:
  - robots
  - wappalyzer

output_modules:
  - csv
  - json

config:
  speculate: False
  excavate: True
"""
    yaml_string_1 = yaml.dump(yaml.safe_load(yaml_string_1), sort_keys=True)
    # preset from yaml
    preset3 = Preset.from_yaml_string(yaml_string_1)
    # yaml to preset
    yaml_string_2 = preset3.to_yaml(sort_keys=True)
    # make sure they're the same
    assert yaml_string_2 == yaml_string_1


def test_preset_cache():
    preset_file = bbot_test_dir / "test_preset.yml"
    yaml_string = """
flags:
  - subdomain-enum

exclude_flags:
  - aggressive
  - slow
"""
    with open(preset_file, "w") as f:
        f.write(yaml_string)

    preset = Preset.from_yaml_file(preset_file)
    assert "subdomain-enum" in preset.flags
    assert "aggressive" in preset.exclude_flags
    assert "slow" in preset.exclude_flags
    from bbot.scanner.preset.preset import _preset_cache

    assert preset_file in _preset_cache

    preset_file.unlink()


def test_preset_scope():

    # test target merging
    scan = Scanner("1.2.3.4", preset=Preset.from_dict({"target": ["evilcorp.com"]}))
    assert set([str(h) for h in scan.preset.target.seeds.hosts]) == {"1.2.3.4", "evilcorp.com"}
    assert set([e.data for e in scan.target]) == {"1.2.3.4", "evilcorp.com"}

    blank_preset = Preset()
    blank_preset = blank_preset.bake()
    assert not blank_preset.target
    assert blank_preset.strict_scope == False

    preset1 = Preset(
        "evilcorp.com",
        "www.evilcorp.ce",
        whitelist=["evilcorp.ce"],
        blacklist=["test.www.evilcorp.ce"],
    )
    preset1_baked = preset1.bake()

    # make sure target logic works as expected
    assert "evilcorp.com" in preset1_baked.target
    assert "asdf.evilcorp.com" in preset1_baked.target
    assert "asdf.www.evilcorp.ce" in preset1_baked.target
    assert not "evilcorp.ce" in preset1_baked.target
    assert "evilcorp.ce" in preset1_baked.whitelist
    assert "test.www.evilcorp.ce" in preset1_baked.blacklist
    assert not "evilcorp.ce" in preset1_baked.blacklist
    assert preset1_baked.in_scope("www.evilcorp.ce")
    assert not preset1_baked.in_scope("evilcorp.com")
    assert not preset1_baked.in_scope("asdf.test.www.evilcorp.ce")

    # test yaml save/load
    yaml1 = preset1.to_yaml(sort_keys=True)
    preset2 = Preset.from_yaml_string(yaml1)
    yaml2 = preset2.to_yaml(sort_keys=True)
    assert yaml1 == yaml2

    # test preset merging
    preset3 = Preset(
        "evilcorp.org",
        whitelist=["evilcorp.de"],
        blacklist=["test.www.evilcorp.de"],
        strict_scope=True,
    )

    preset1.merge(preset3)

    preset1_baked = preset1.bake()

    # targets should be merged
    assert "evilcorp.com" in preset1_baked.target
    assert "www.evilcorp.ce" in preset1_baked.target
    assert "evilcorp.org" in preset1_baked.target
    # strict scope is enabled
    assert not "asdf.evilcorp.com" in preset1_baked.target
    assert not "asdf.www.evilcorp.ce" in preset1_baked.target
    assert "evilcorp.ce" in preset1_baked.whitelist
    assert "evilcorp.de" in preset1_baked.whitelist
    assert not "asdf.evilcorp.de" in preset1_baked.whitelist
    assert not "asdf.evilcorp.ce" in preset1_baked.whitelist
    # blacklist should be merged, strict scope does not apply
    assert "asdf.test.www.evilcorp.ce" in preset1_baked.blacklist
    assert "asdf.test.www.evilcorp.de" in preset1_baked.blacklist
    assert not "asdf.test.www.evilcorp.org" in preset1_baked.blacklist
    # only the base domain of evilcorp.de should be in scope
    assert not preset1_baked.in_scope("evilcorp.com")
    assert not preset1_baked.in_scope("evilcorp.org")
    assert preset1_baked.in_scope("evilcorp.de")
    assert not preset1_baked.in_scope("asdf.evilcorp.de")
    assert not preset1_baked.in_scope("evilcorp.com")
    assert not preset1_baked.in_scope("asdf.test.www.evilcorp.ce")

    preset4 = Preset(output_modules="neo4j")
    set(preset1.output_modules) == {"python", "csv", "txt", "json", "stdout"}
    preset1.merge(preset4)
    set(preset1.output_modules) == {"python", "csv", "txt", "json", "stdout", "neo4j"}

    # test preset merging + whitelist

    preset_nowhitelist = Preset("evilcorp.com", name="nowhitelist")
    preset_whitelist = Preset(
        "evilcorp.org",
        name="whitelist",
        whitelist=["1.2.3.4/24", "http://evilcorp.net"],
        blacklist=["evilcorp.co.uk:443", "bob@evilcorp.co.uk"],
        config={"modules": {"secretsdb": {"api_key": "deadbeef", "otherthing": "asdf"}}},
    )

    preset_nowhitelist_baked = preset_nowhitelist.bake()
    preset_whitelist_baked = preset_whitelist.bake()

    assert preset_nowhitelist_baked.to_dict(include_target=True) == {
        "target": ["evilcorp.com"],
    }
    assert preset_whitelist_baked.to_dict(include_target=True) == {
        "target": ["evilcorp.org"],
        "whitelist": ["1.2.3.0/24", "evilcorp.net"],
        "blacklist": ["evilcorp.co.uk"],
        "config": {"modules": {"secretsdb": {"api_key": "deadbeef", "otherthing": "asdf"}}},
    }
    assert preset_whitelist_baked.to_dict(include_target=True, redact_secrets=True) == {
        "target": ["evilcorp.org"],
        "whitelist": ["1.2.3.0/24", "evilcorp.net"],
        "blacklist": ["evilcorp.co.uk"],
        "config": {"modules": {"secretsdb": {"otherthing": "asdf"}}},
    }

    assert preset_nowhitelist_baked.in_scope("www.evilcorp.com")
    assert not preset_nowhitelist_baked.in_scope("www.evilcorp.de")
    assert not preset_nowhitelist_baked.in_scope("1.2.3.4/24")

    assert "www.evilcorp.org" in preset_whitelist_baked.target
    assert "1.2.3.4" in preset_whitelist_baked.whitelist
    assert not preset_whitelist_baked.in_scope("www.evilcorp.org")
    assert not preset_whitelist_baked.in_scope("www.evilcorp.de")
    assert not preset_whitelist_baked.whitelisted("www.evilcorp.org")
    assert not preset_whitelist_baked.whitelisted("www.evilcorp.de")
    assert preset_whitelist_baked.in_scope("1.2.3.4")
    assert preset_whitelist_baked.in_scope("1.2.3.4/28")
    assert preset_whitelist_baked.in_scope("1.2.3.4/24")
    assert preset_whitelist_baked.whitelisted("1.2.3.4")
    assert preset_whitelist_baked.whitelisted("1.2.3.4/28")
    assert preset_whitelist_baked.whitelisted("1.2.3.4/24")

    assert set([e.data for e in preset_nowhitelist_baked.target]) == {"evilcorp.com"}
    assert set([e.data for e in preset_whitelist_baked.target]) == {"evilcorp.org"}
    assert set([e.data for e in preset_nowhitelist_baked.whitelist]) == {"evilcorp.com"}
    assert set([e.data for e in preset_whitelist_baked.whitelist]) == {"1.2.3.0/24", "evilcorp.net"}

    preset_nowhitelist.merge(preset_whitelist)
    preset_nowhitelist_baked = preset_nowhitelist.bake()
    assert set([e.data for e in preset_nowhitelist_baked.target]) == {"evilcorp.com", "evilcorp.org"}
    assert set([e.data for e in preset_nowhitelist_baked.whitelist]) == {"1.2.3.0/24", "evilcorp.net"}
    assert "www.evilcorp.org" in preset_nowhitelist_baked.target
    assert "www.evilcorp.com" in preset_nowhitelist_baked.target
    assert "1.2.3.4" in preset_nowhitelist_baked.whitelist
    assert not preset_nowhitelist_baked.in_scope("www.evilcorp.org")
    assert not preset_nowhitelist_baked.in_scope("www.evilcorp.com")
    assert not preset_nowhitelist_baked.whitelisted("www.evilcorp.org")
    assert not preset_nowhitelist_baked.whitelisted("www.evilcorp.com")
    assert preset_nowhitelist_baked.in_scope("1.2.3.4")

    preset_nowhitelist = Preset("evilcorp.com")
    preset_whitelist = Preset("evilcorp.org", whitelist=["1.2.3.4/24"])
    preset_whitelist.merge(preset_nowhitelist)
    preset_whitelist_baked = preset_whitelist.bake()
    assert set([e.data for e in preset_whitelist_baked.target]) == {"evilcorp.com", "evilcorp.org"}
    assert set([e.data for e in preset_whitelist_baked.whitelist]) == {"1.2.3.0/24"}
    assert "www.evilcorp.org" in preset_whitelist_baked.target
    assert "www.evilcorp.com" in preset_whitelist_baked.target
    assert "1.2.3.4" in preset_whitelist_baked.whitelist
    assert not preset_whitelist_baked.in_scope("www.evilcorp.org")
    assert not preset_whitelist_baked.in_scope("www.evilcorp.com")
    assert not preset_whitelist_baked.whitelisted("www.evilcorp.org")
    assert not preset_whitelist_baked.whitelisted("www.evilcorp.com")
    assert preset_whitelist_baked.in_scope("1.2.3.4")

    preset_nowhitelist1 = Preset("evilcorp.com")
    preset_nowhitelist2 = Preset("evilcorp.de")
    preset_nowhitelist1_baked = preset_nowhitelist1.bake()
    preset_nowhitelist2_baked = preset_nowhitelist2.bake()
    assert set([e.data for e in preset_nowhitelist1_baked.target]) == {"evilcorp.com"}
    assert set([e.data for e in preset_nowhitelist2_baked.target]) == {"evilcorp.de"}
    assert set([e.data for e in preset_nowhitelist1_baked.whitelist]) == {"evilcorp.com"}
    assert set([e.data for e in preset_nowhitelist2_baked.whitelist]) == {"evilcorp.de"}
    preset_nowhitelist1.merge(preset_nowhitelist2)
    preset_nowhitelist1_baked = preset_nowhitelist1.bake()
    assert set([e.data for e in preset_nowhitelist1_baked.target]) == {"evilcorp.com", "evilcorp.de"}
    assert set([e.data for e in preset_nowhitelist2_baked.target]) == {"evilcorp.de"}
    assert set([e.data for e in preset_nowhitelist1_baked.whitelist]) == {"evilcorp.com", "evilcorp.de"}
    assert set([e.data for e in preset_nowhitelist2_baked.whitelist]) == {"evilcorp.de"}
    assert "www.evilcorp.com" in preset_nowhitelist1_baked.target
    assert "www.evilcorp.de" in preset_nowhitelist1_baked.target
    assert "www.evilcorp.com" in preset_nowhitelist1_baked.target.seeds
    assert "www.evilcorp.de" in preset_nowhitelist1_baked.target.seeds
    assert "www.evilcorp.com" in preset_nowhitelist1_baked.whitelist
    assert "www.evilcorp.de" in preset_nowhitelist1_baked.whitelist
    assert preset_nowhitelist1_baked.whitelisted("www.evilcorp.com")
    assert preset_nowhitelist1_baked.whitelisted("www.evilcorp.de")
    assert not preset_nowhitelist1_baked.whitelisted("1.2.3.4")
    assert preset_nowhitelist1_baked.in_scope("www.evilcorp.com")
    assert preset_nowhitelist1_baked.in_scope("www.evilcorp.de")
    assert not preset_nowhitelist1_baked.in_scope("1.2.3.4")

    preset_nowhitelist1 = Preset("evilcorp.com")
    preset_nowhitelist2 = Preset("evilcorp.de")
    preset_nowhitelist2.merge(preset_nowhitelist1)
    preset_nowhitelist1_baked = preset_nowhitelist1.bake()
    preset_nowhitelist2_baked = preset_nowhitelist2.bake()
    assert set([e.data for e in preset_nowhitelist1_baked.target]) == {"evilcorp.com"}
    assert set([e.data for e in preset_nowhitelist2_baked.target]) == {"evilcorp.com", "evilcorp.de"}
    assert set([e.data for e in preset_nowhitelist1_baked.whitelist]) == {"evilcorp.com"}
    assert set([e.data for e in preset_nowhitelist2_baked.whitelist]) == {"evilcorp.com", "evilcorp.de"}


@pytest.mark.asyncio
async def test_preset_logging():

    scan = Scanner()

    # test individual verbosity levels
    original_log_level = CORE.logger.log_level
    assert original_log_level == logging.DEBUG

    try:
        silent_preset = Preset(silent=True)
        assert silent_preset.silent == True
        assert silent_preset.debug == False
        assert silent_preset.verbose == False
        assert original_log_level == CORE.logger.log_level
        debug_preset = Preset(debug=True)
        assert debug_preset.silent == False
        assert debug_preset.debug == True
        assert debug_preset.verbose == False
        assert original_log_level == CORE.logger.log_level
        verbose_preset = Preset(verbose=True)
        assert verbose_preset.silent == False
        assert verbose_preset.debug == False
        assert verbose_preset.verbose == True
        assert original_log_level == CORE.logger.log_level

        # test conflicting verbosity levels
        silent_and_verbose = Preset(silent=True, verbose=True)
        assert silent_and_verbose.silent == True
        assert silent_and_verbose.debug == False
        assert silent_and_verbose.verbose == True
        baked = silent_and_verbose.bake()
        assert baked.silent == True
        assert baked.debug == False
        assert baked.verbose == False
        assert baked.core.logger.log_level == original_log_level
        baked = silent_and_verbose.bake(scan=scan)
        assert baked.core.logger.log_level == logging.CRITICAL
        assert CORE.logger.log_level == logging.CRITICAL

        CORE.logger.log_level = original_log_level
        assert CORE.logger.log_level == original_log_level

        silent_and_debug = Preset(silent=True, debug=True)
        assert silent_and_debug.silent == True
        assert silent_and_debug.debug == True
        assert silent_and_debug.verbose == False
        baked = silent_and_debug.bake()
        assert baked.silent == True
        assert baked.debug == False
        assert baked.verbose == False
        assert baked.core.logger.log_level == original_log_level
        baked = silent_and_debug.bake(scan=scan)
        assert baked.core.logger.log_level == logging.CRITICAL
        assert CORE.logger.log_level == logging.CRITICAL

        CORE.logger.log_level = original_log_level
        assert CORE.logger.log_level == original_log_level

        debug_and_verbose = Preset(verbose=True, debug=True)
        assert debug_and_verbose.silent == False
        assert debug_and_verbose.debug == True
        assert debug_and_verbose.verbose == True
        baked = debug_and_verbose.bake()
        assert baked.silent == False
        assert baked.debug == True
        assert baked.verbose == False
        assert baked.core.logger.log_level == original_log_level
        baked = debug_and_verbose.bake(scan=scan)
        assert baked.core.logger.log_level == logging.DEBUG
        assert CORE.logger.log_level == logging.DEBUG

        CORE.logger.log_level = original_log_level
        assert CORE.logger.log_level == original_log_level

        all_preset = Preset(verbose=True, debug=True, silent=True)
        assert all_preset.silent == True
        assert all_preset.debug == True
        assert all_preset.verbose == True
        baked = all_preset.bake()
        assert baked.silent == True
        assert baked.debug == False
        assert baked.verbose == False
        assert baked.core.logger.log_level == original_log_level
        baked = all_preset.bake(scan=scan)
        assert baked.core.logger.log_level == logging.CRITICAL
        assert CORE.logger.log_level == logging.CRITICAL

        CORE.logger.log_level = original_log_level
        assert CORE.logger.log_level == original_log_level

        # defaults
        preset = Preset().bake()
        assert preset.core.logger.log_level == original_log_level
        assert CORE.logger.log_level == original_log_level

    finally:
        CORE.logger.log_level = original_log_level
        assert CORE.logger.log_level == original_log_level
        await scan._cleanup()


def test_preset_module_resolution(clean_default_config):
    preset = Preset().bake()
    sslcert_preloaded = preset.preloaded_module("sslcert")
    wayback_preloaded = preset.preloaded_module("wayback")
    wappalyzer_preloaded = preset.preloaded_module("wappalyzer")
    sslcert_flags = sslcert_preloaded.get("flags", [])
    wayback_flags = wayback_preloaded.get("flags", [])
    wappalyzer_flags = wappalyzer_preloaded.get("flags", [])
    assert "active" in sslcert_flags
    assert "passive" in wayback_flags
    assert "active" in wappalyzer_flags
    assert "subdomain-enum" in sslcert_flags
    assert "subdomain-enum" in wayback_flags
    assert "httpx" in wappalyzer_preloaded["deps"]["modules"]

    # make sure we have the expected defaults
    assert not preset.scan_modules
    assert set(preset.output_modules) == {"python", "csv", "txt", "json"}
    assert set(preset.internal_modules) == {"aggregate", "excavate", "speculate", "cloudcheck", "dnsresolve"}
    assert preset.modules == set(preset.output_modules).union(set(preset.internal_modules))

    # make sure dependency resolution works as expected
    preset = Preset(modules=["wappalyzer"]).bake()
    assert set(preset.scan_modules) == {"wappalyzer", "httpx"}

    # make sure flags work as expected
    preset = Preset(flags=["subdomain-enum"]).bake()
    assert preset.flags == {"subdomain-enum"}
    assert "sslcert" in preset.modules
    assert "wayback" in preset.modules
    assert "sslcert" in preset.scan_modules
    assert "wayback" in preset.scan_modules

    # flag + module exclusions
    preset = Preset(flags=["subdomain-enum"], exclude_modules=["sslcert"]).bake()
    assert "sslcert" not in preset.modules
    assert "wayback" in preset.modules
    assert "sslcert" not in preset.scan_modules
    assert "wayback" in preset.scan_modules

    # flag + flag exclusions
    preset = Preset(flags=["subdomain-enum"], exclude_flags=["active"]).bake()
    assert "sslcert" not in preset.modules
    assert "wayback" in preset.modules
    assert "sslcert" not in preset.scan_modules
    assert "wayback" in preset.scan_modules

    # flag + flag requirements
    preset = Preset(flags=["subdomain-enum"], require_flags=["passive"]).bake()
    assert "sslcert" not in preset.modules
    assert "wayback" in preset.modules
    assert "sslcert" not in preset.scan_modules
    assert "wayback" in preset.scan_modules

    # normal module enableement
    preset = Preset(modules=["sslcert", "wappalyzer", "wayback"]).bake()
    assert set(preset.scan_modules) == {"sslcert", "wappalyzer", "wayback", "httpx"}

    # modules + flag exclusions
    preset = Preset(exclude_flags=["active"], modules=["sslcert", "wappalyzer", "wayback"]).bake()
    assert set(preset.scan_modules) == {"wayback"}

    # modules + flag requirements
    preset = Preset(require_flags=["passive"], modules=["sslcert", "wappalyzer", "wayback"]).bake()
    assert set(preset.scan_modules) == {"wayback"}

    # modules + module exclusions
    preset = Preset(exclude_modules=["sslcert"], modules=["sslcert", "wappalyzer", "wayback"]).bake()
    baked_preset = preset.bake()
    assert baked_preset.modules == {
        "wayback",
        "cloudcheck",
        "python",
        "json",
        "speculate",
        "dnsresolve",
        "aggregate",
        "excavate",
        "txt",
        "httpx",
        "csv",
        "wappalyzer",
    }


@pytest.mark.asyncio
async def test_preset_module_loader():
    custom_module_dir = bbot_test_dir / "custom_module_dir"
    custom_module_dir_2 = custom_module_dir / "asdf"
    custom_output_module_dir = custom_module_dir / "output"
    custom_internal_module_dir = custom_module_dir / "internal"
    for d in [custom_module_dir, custom_module_dir_2, custom_output_module_dir, custom_internal_module_dir]:
        d.mkdir(parents=True, exist_ok=True)
        assert d.is_dir()
    custom_module_1 = custom_module_dir / "testmodule1.py"
    with open(custom_module_1, "w") as f:
        f.write(
            """
from bbot.modules.base import BaseModule

class TestModule1(BaseModule):
    watched_events = ["URL", "HTTP_RESPONSE"]
    produced_events = ["VULNERABILITY"]
"""
        )

    custom_module_2 = custom_output_module_dir / "testmodule2.py"
    with open(custom_module_2, "w") as f:
        f.write(
            """
from bbot.modules.output.base import BaseOutputModule

class TestModule2(BaseOutputModule):
    pass
"""
        )

    custom_module_3 = custom_internal_module_dir / "testmodule3.py"
    with open(custom_module_3, "w") as f:
        f.write(
            """
from bbot.modules.internal.base import BaseInternalModule

class TestModule3(BaseInternalModule):
    pass
"""
        )

    custom_module_4 = custom_module_dir_2 / "testmodule4.py"
    with open(custom_module_4, "w") as f:
        f.write(
            """
from bbot.modules.base import BaseModule

class TestModule4(BaseModule):
    watched_events = ["TECHNOLOGY"]
    produced_events = ["FINDING"]
"""
        )

    assert custom_module_1.is_file()
    assert custom_module_2.is_file()
    assert custom_module_3.is_file()
    assert custom_module_4.is_file()

    preset = Preset()
    preset.module_loader.save_preload_cache()
    assert preset.module_loader.preload_cache_file.is_file()

    # at this point, core modules should be loaded, but not custom ones
    assert "wappalyzer" in preset.module_loader.preloaded()
    assert "testmodule1" not in preset.module_loader.preloaded()

    import pickle

    with open(preset.module_loader.preload_cache_file, "rb") as f:
        preloaded = pickle.load(f)
    assert "wappalyzer" in preloaded
    assert "testmodule1" not in preloaded

    # add custom module dir
    preset.module_dirs = [str(custom_module_dir)]
    assert custom_module_dir in preset.module_dirs
    assert custom_module_dir_2 in preset.module_dirs
    assert custom_output_module_dir in preset.module_dirs
    assert custom_internal_module_dir in preset.module_dirs

    # now our custom modules should be loaded
    assert "wappalyzer" in preset.module_loader.preloaded()
    assert "testmodule1" in preset.module_loader.preloaded()
    assert "testmodule2" in preset.module_loader.preloaded()
    assert "testmodule3" in preset.module_loader.preloaded()
    assert "testmodule4" in preset.module_loader.preloaded()

    preset.module_loader.save_preload_cache()
    with open(preset.module_loader.preload_cache_file, "rb") as f:
        preloaded = pickle.load(f)
    assert "wappalyzer" in preloaded
    assert "testmodule1" in preloaded
    assert "testmodule2" in preloaded
    assert "testmodule3" in preloaded
    assert "testmodule4" in preloaded

    # since module loader is shared across all presets, a new preset should now also have our custom modules
    preset2 = Preset()
    assert "wappalyzer" in preset2.module_loader.preloaded()
    assert "testmodule1" in preset2.module_loader.preloaded()
    assert "testmodule2" in preset2.module_loader.preloaded()
    assert "testmodule3" in preset2.module_loader.preloaded()
    assert "testmodule4" in preset2.module_loader.preloaded()

    # reset module_loader
    preset2.module_loader.__init__()

    # custom module dir via preset
    custom_module_dir_3 = bbot_test_dir / "custom_module_dir_3"
    custom_module_dir_3.mkdir(exist_ok=True, parents=True)
    custom_module_5 = custom_module_dir_3 / "testmodule5.py"
    with open(custom_module_5, "w") as f:
        f.write(
            """
from bbot.modules.base import BaseModule

class TestModule5(BaseModule):
    watched_events = ["TECHNOLOGY"]
    produced_events = ["FINDING"]
"""
        )

    preset = Preset.from_yaml_string(
        f"""
modules:
  - testmodule5
"""
    )
    # should fail
    with pytest.raises(ValidationError):
        scan = Scanner(preset=preset)

    preset = Preset.from_yaml_string(
        f"""
module_dirs:
  - {custom_module_dir_3}
modules:
  - testmodule5
"""
    )
    scan = Scanner(preset=preset)
    await scan._prep()
    assert "testmodule5" in scan.modules


def test_preset_include():

    # test recursive preset inclusion

    custom_preset_dir_1 = bbot_test_dir / "custom_preset_dir"
    custom_preset_dir_2 = custom_preset_dir_1 / "preset_subdir"
    custom_preset_dir_3 = custom_preset_dir_2 / "subsubdir"
    custom_preset_dir_4 = Path("/tmp/.bbot_preset_test")
    custom_preset_dir_5 = custom_preset_dir_4 / "subdir"
    mkdir(custom_preset_dir_1)
    mkdir(custom_preset_dir_2)
    mkdir(custom_preset_dir_3)
    mkdir(custom_preset_dir_4)
    mkdir(custom_preset_dir_5)

    preset_file = custom_preset_dir_1 / "preset1.yml"
    with open(preset_file, "w") as f:
        f.write(
            """
include:
  - preset2

config:
  modules:
    testpreset1:
      test: asdf
"""
        )

    preset_file = custom_preset_dir_2 / "preset2.yml"
    with open(preset_file, "w") as f:
        f.write(
            """
include:
  - preset3

config:
  modules:
    testpreset2:
      test: fdsa
"""
        )

    preset_file = custom_preset_dir_3 / "preset3.yml"
    with open(preset_file, "w") as f:
        f.write(
            f"""
include:
  # uh oh
  - preset1
  - {custom_preset_dir_4}/preset4

config:
  modules:
    testpreset3:
      test: qwerty
"""
        )

    preset_file = custom_preset_dir_4 / "preset4.yml"
    with open(preset_file, "w") as f:
        f.write(
            """
include:
  - preset5

config:
  modules:
    testpreset4:
      test: zxcv
"""
        )

    preset_file = custom_preset_dir_5 / "preset5.yml"
    with open(preset_file, "w") as f:
        f.write(
            """
config:
  modules:
    testpreset5:
      test: hjkl
"""
        )

    # with include=
    preset = Preset(include=[str(custom_preset_dir_1 / "preset1")])
    assert preset.config.modules.testpreset1.test == "asdf"
    assert preset.config.modules.testpreset2.test == "fdsa"
    assert preset.config.modules.testpreset3.test == "qwerty"
    assert preset.config.modules.testpreset4.test == "zxcv"
    assert preset.config.modules.testpreset5.test == "hjkl"

    # same thing but with presets= (an alias to include)
    preset = Preset(presets=[str(custom_preset_dir_1 / "preset1")])
    assert preset.config.modules.testpreset1.test == "asdf"
    assert preset.config.modules.testpreset2.test == "fdsa"
    assert preset.config.modules.testpreset3.test == "qwerty"
    assert preset.config.modules.testpreset4.test == "zxcv"
    assert preset.config.modules.testpreset5.test == "hjkl"

    # can't use both include= and presets= at the same time
    with pytest.raises(ValueError):
        preset = Preset(presets=["subdomain-enum"], include=["dirbust-light"])


@pytest.mark.asyncio
async def test_preset_conditions():
    custom_preset_dir_1 = bbot_test_dir / "custom_preset_dir"
    custom_preset_dir_2 = custom_preset_dir_1 / "preset_subdir"
    mkdir(custom_preset_dir_1)
    mkdir(custom_preset_dir_2)

    preset_file_1 = custom_preset_dir_1 / "preset_condition_1.yml"
    with open(preset_file_1, "w") as f:
        f.write(
            """
include:
  - preset_condition_2
"""
        )

    preset_file_2 = custom_preset_dir_2 / "preset_condition_2.yml"
    with open(preset_file_2, "w") as f:
        f.write(
            """
conditions:
  - |
    {% if config.web.spider_distance == 3 and config.web.spider_depth == 4 %}
      {{ abort("web spider is too aggressive") }}
    {% endif %}
"""
        )

    preset = Preset(include=[preset_file_1])
    assert preset.conditions

    scan = Scanner(preset=preset)
    assert scan.preset.conditions

    await scan._cleanup()

    preset2 = Preset(config={"web": {"spider_distance": 3, "spider_depth": 4}})
    preset.merge(preset2)

    with pytest.raises(PresetAbortError):
        Scanner(preset=preset)


def test_preset_module_disablement(clean_default_config):
    # internal module disablement
    preset = Preset().bake()
    assert "speculate" in preset.internal_modules
    assert "excavate" in preset.internal_modules
    assert "aggregate" in preset.internal_modules
    preset = Preset(config={"speculate": False}).bake()
    assert "speculate" not in preset.internal_modules
    assert "excavate" in preset.internal_modules
    assert "aggregate" in preset.internal_modules
    preset = Preset(exclude_modules=["speculate", "excavate"]).bake()
    assert "speculate" not in preset.internal_modules
    assert "excavate" not in preset.internal_modules
    assert "aggregate" in preset.internal_modules

    # internal module disablement
    preset = Preset().bake()
    assert set(preset.output_modules) == {"python", "txt", "csv", "json"}
    preset = Preset(exclude_modules=["txt", "csv"]).bake()
    assert set(preset.output_modules) == {"python", "json"}
    preset = Preset(output_modules=["json"]).bake()
    assert set(preset.output_modules) == {"json"}


def test_preset_require_exclude():

    def get_module_flags(p):
        for m in p.scan_modules:
            preloaded = p.preloaded_module(m)
            yield m, preloaded.get("flags", [])

    # enable by flag, no exclusions/requirements
    preset = Preset(flags=["subdomain-enum"]).bake()
    assert len(preset.modules) > 25
    module_flags = list(get_module_flags(preset))
    dnsbrute_flags = preset.preloaded_module("dnsbrute").get("flags", [])
    assert "subdomain-enum" in dnsbrute_flags
    assert "active" in dnsbrute_flags
    assert not "passive" in dnsbrute_flags
    assert "aggressive" in dnsbrute_flags
    assert not "safe" in dnsbrute_flags
    assert "dnsbrute" in [x[0] for x in module_flags]
    assert "certspotter" in [x[0] for x in module_flags]
    assert "c99" in [x[0] for x in module_flags]
    assert any("passive" in flags for module, flags in module_flags)
    assert any("active" in flags for module, flags in module_flags)
    assert any("safe" in flags for module, flags in module_flags)
    assert any("aggressive" in flags for module, flags in module_flags)

    # enable by flag, one required flag
    preset = Preset(flags=["subdomain-enum"], require_flags=["passive"]).bake()
    assert len(preset.modules) > 25
    module_flags = list(get_module_flags(preset))
    assert "chaos" in [x[0] for x in module_flags]
    assert not "httpx" in [x[0] for x in module_flags]
    assert all("passive" in flags for module, flags in module_flags)
    assert not any("active" in flags for module, flags in module_flags)
    assert any("safe" in flags for module, flags in module_flags)
    assert any("aggressive" in flags for module, flags in module_flags)

    # enable by flag, one excluded flag
    preset = Preset(flags=["subdomain-enum"], exclude_flags=["active"]).bake()
    assert len(preset.modules) > 25
    module_flags = list(get_module_flags(preset))
    assert "chaos" in [x[0] for x in module_flags]
    assert not "httpx" in [x[0] for x in module_flags]
    assert all("passive" in flags for module, flags in module_flags)
    assert not any("active" in flags for module, flags in module_flags)
    assert any("safe" in flags for module, flags in module_flags)
    assert any("aggressive" in flags for module, flags in module_flags)

    # enable by flag, one excluded module
    preset = Preset(flags=["subdomain-enum"], exclude_modules=["dnsbrute"]).bake()
    assert len(preset.modules) > 25
    module_flags = list(get_module_flags(preset))
    assert not "dnsbrute" in [x[0] for x in module_flags]
    assert "httpx" in [x[0] for x in module_flags]
    assert any("passive" in flags for module, flags in module_flags)
    assert any("active" in flags for module, flags in module_flags)
    assert any("safe" in flags for module, flags in module_flags)
    assert any("aggressive" in flags for module, flags in module_flags)

    # enable by flag, multiple required flags
    preset = Preset(flags=["subdomain-enum"], require_flags=["safe", "passive"]).bake()
    assert len(preset.modules) > 25
    module_flags = list(get_module_flags(preset))
    assert not "dnsbrute" in [x[0] for x in module_flags]
    assert all("passive" in flags and "safe" in flags for module, flags in module_flags)
    assert all("active" not in flags and "aggressive" not in flags for module, flags in module_flags)
    assert not any("active" in flags for module, flags in module_flags)
    assert not any("aggressive" in flags for module, flags in module_flags)

    # enable by flag, multiple excluded flags
    preset = Preset(flags=["subdomain-enum"], exclude_flags=["aggressive", "active"]).bake()
    assert len(preset.modules) > 25
    module_flags = list(get_module_flags(preset))
    assert not "dnsbrute" in [x[0] for x in module_flags]
    assert all("passive" in flags and "safe" in flags for module, flags in module_flags)
    assert all("active" not in flags and "aggressive" not in flags for module, flags in module_flags)
    assert not any("active" in flags for module, flags in module_flags)
    assert not any("aggressive" in flags for module, flags in module_flags)

    # enable by flag, multiple excluded modules
    preset = Preset(flags=["subdomain-enum"], exclude_modules=["dnsbrute", "c99"]).bake()
    assert len(preset.modules) > 25
    module_flags = list(get_module_flags(preset))
    assert not "dnsbrute" in [x[0] for x in module_flags]
    assert "certspotter" in [x[0] for x in module_flags]
    assert not "c99" in [x[0] for x in module_flags]
    assert any("passive" in flags for module, flags in module_flags)
    assert any("active" in flags for module, flags in module_flags)
    assert any("safe" in flags for module, flags in module_flags)
    assert any("aggressive" in flags for module, flags in module_flags)


@pytest.mark.asyncio
async def test_preset_output_dir():
    output_dir = bbot_test_dir / "preset_output_dir"
    preset = Preset.from_yaml_string(
        f"""
output_dir: {output_dir}
scan_name: bbot_test
"""
    )
    scan = Scanner(preset=preset)
    await scan.async_start_without_generator()
    scan_dir = output_dir / "bbot_test"
    assert scan_dir.is_dir()
    output_file = scan_dir / "output.txt"
    assert output_file.is_file()

    shutil.rmtree(output_dir, ignore_errors=True)



from ..module_tests.base import ModuleTestBase


class TestSubdomainEnum(ModuleTestBase):
    targets = ["blacklanternsecurity.com"]
    modules_overrides = []
    config_overrides = {"dns": {"minimal": False}, "scope": {"report_distance": 10}}
    dedup_strategy = "highest_parent"

    txt = [
        "www.blacklanternsecurity.com",
        "asdf.www.blacklanternsecurity.com",
        "test.asdf.www.blacklanternsecurity.com",
        "api.test.asdf.www.blacklanternsecurity.com",
    ]

    async def setup_after_prep(self, module_test):
        dns_mock = {
            "evilcorp.com": {"A": ["127.0.0.6"]},
            "blacklanternsecurity.com": {"A": ["127.0.0.5"]},
            "www.blacklanternsecurity.com": {"A": ["127.0.0.5"]},
            "asdf.www.blacklanternsecurity.com": {"A": ["127.0.0.5"]},
            "test.asdf.www.blacklanternsecurity.com": {"A": ["127.0.0.5"]},
            "api.test.asdf.www.blacklanternsecurity.com": {"A": ["127.0.0.5"]},
        }
        if self.txt:
            dns_mock["blacklanternsecurity.com"]["TXT"] = self.txt
        await module_test.mock_dns(dns_mock)

        # load subdomain enum template as module
        from bbot.modules.templates.subdomain_enum import subdomain_enum

        subdomain_enum_module = subdomain_enum(module_test.scan)

        self.queries = []

        async def mock_query(query):
            self.queries.append(query)

        subdomain_enum_module.query = mock_query
        subdomain_enum_module.dedup_strategy = self.dedup_strategy
        module_test.scan.modules["subdomain_enum"] = subdomain_enum_module

    def check(self, module_test, events):
        in_scope_dns_names = [e for e in events if e.type == "DNS_NAME" and e.scope_distance == 0]
        assert len(in_scope_dns_names) == 5
        assert 1 == len([e for e in in_scope_dns_names if e.data == "blacklanternsecurity.com"])
        assert 1 == len([e for e in in_scope_dns_names if e.data == "www.blacklanternsecurity.com"])
        assert 1 == len([e for e in in_scope_dns_names if e.data == "asdf.www.blacklanternsecurity.com"])
        assert 1 == len([e for e in in_scope_dns_names if e.data == "test.asdf.www.blacklanternsecurity.com"])
        assert 1 == len([e for e in in_scope_dns_names if e.data == "api.test.asdf.www.blacklanternsecurity.com"])
        assert len(self.queries) == 1
        assert self.queries[0] == "blacklanternsecurity.com"


class TestSubdomainEnumHighestParent(TestSubdomainEnum):
    targets = ["api.test.asdf.www.blacklanternsecurity.com", "evilcorp.com"]
    whitelist = ["www.blacklanternsecurity.com"]
    modules_overrides = ["speculate"]
    dedup_strategy = "highest_parent"
    txt = None

    def check(self, module_test, events):
        in_scope_dns_names = [e for e in events if e.type == "DNS_NAME" and e.scope_distance == 0]
        distance_1_dns_names = [e for e in events if e.type == "DNS_NAME" and e.scope_distance == 1]
        assert len(in_scope_dns_names) == 4
        assert 1 == len([e for e in in_scope_dns_names if e.data == "www.blacklanternsecurity.com"])
        assert 1 == len([e for e in in_scope_dns_names if e.data == "asdf.www.blacklanternsecurity.com"])
        assert 1 == len([e for e in in_scope_dns_names if e.data == "test.asdf.www.blacklanternsecurity.com"])
        assert 1 == len([e for e in in_scope_dns_names if e.data == "api.test.asdf.www.blacklanternsecurity.com"])
        assert len(distance_1_dns_names) == 2
        assert 1 == len([e for e in distance_1_dns_names if e.data == "evilcorp.com"])
        assert 1 == len([e for e in distance_1_dns_names if e.data == "blacklanternsecurity.com"])
        assert len(self.queries) == 1
        assert self.queries[0] == "www.blacklanternsecurity.com"


class TestSubdomainEnumLowestParent(TestSubdomainEnumHighestParent):
    dedup_strategy = "lowest_parent"

    def check(self, module_test, events):
        assert set(self.queries) == {
            "test.asdf.www.blacklanternsecurity.com",
            "asdf.www.blacklanternsecurity.com",
            "www.blacklanternsecurity.com",
        }


class TestSubdomainEnumWildcardBaseline(ModuleTestBase):
    # oh walmart.cn why are you like this
    targets = ["www.walmart.cn"]
    whitelist = ["walmart.cn"]
    modules_overrides = []
    config_overrides = {"dns": {"minimal": False}, "scope": {"report_distance": 10}, "omit_event_types": []}
    dedup_strategy = "highest_parent"

    dns_mock_data = {
        "walmart.cn": {"A": ["127.0.0.1"]},
        "www.walmart.cn": {"A": ["127.0.0.1"]},
        "test.walmart.cn": {"A": ["127.0.0.1"]},
    }

    async def setup_before_prep(self, module_test):
        await module_test.mock_dns(self.dns_mock_data)
        self.queries = []

        async def mock_query(query):
            self.queries.append(query)
            return ["walmart.cn", "www.walmart.cn", "test.walmart.cn", "asdf.walmart.cn"]

        # load subdomain enum template as module
        from bbot.modules.templates.subdomain_enum import subdomain_enum

        subdomain_enum_module = subdomain_enum(module_test.scan)

        subdomain_enum_module.query = mock_query
        subdomain_enum_module._name = "subdomain_enum"
        subdomain_enum_module.dedup_strategy = self.dedup_strategy
        module_test.scan.modules["subdomain_enum"] = subdomain_enum_module

    def check(self, module_test, events):
        assert self.queries == ["walmart.cn"]
        assert len(events) == 7
        assert 2 == len(
            [
                e
                for e in events
                if e.type == "IP_ADDRESS" and e.data == "127.0.0.1" and str(e.module) == "A" and e.scope_distance == 1
            ]
        )
        assert 1 == len(
            [
                e
                for e in events
                if e.type == "DNS_NAME"
                and e.data == "www.walmart.cn"
                and str(e.module) == "TARGET"
                and e.scope_distance == 0
            ]
        )
        assert 1 == len(
            [
                e
                for e in events
                if e.type == "DNS_NAME"
                and e.data == "test.walmart.cn"
                and str(e.module) == "subdomain_enum"
                and e.scope_distance == 0
            ]
        )
        assert 1 == len(
            [
                e
                for e in events
                if e.type == "DNS_NAME_UNRESOLVED"
                and e.data == "asdf.walmart.cn"
                and str(e.module) == "subdomain_enum"
                and e.scope_distance == 0
            ]
        )


class TestSubdomainEnumWildcardDefense(TestSubdomainEnumWildcardBaseline):
    # oh walmart.cn why are you like this
    targets = ["walmart.cn"]
    modules_overrides = []
    config_overrides = {"dns": {"minimal": False}, "scope": {"report_distance": 10}}
    dedup_strategy = "highest_parent"

    dns_mock_data = {
        "walmart.cn": {"A": ["127.0.0.2"], "TXT": ["asdf.walmart.cn"]},
    }

    async def setup_after_prep(self, module_test):
        # simulate wildcard
        custom_lookup = """
def custom_lookup(query, rdtype):
    import random
    if rdtype == "A" and query.endswith(".walmart.cn"):
        ip = ".".join([str(random.randint(0,256)) for _ in range(4)])
        return {ip}
"""
        await module_test.mock_dns(self.dns_mock_data, custom_lookup_fn=custom_lookup)

    def check(self, module_test, events):
        # no subdomain enum should happen on this domain!
        assert self.queries == []
        assert len(events) == 7
        assert 2 == len(
            [e for e in events if e.type == "IP_ADDRESS" and str(e.module) == "A" and e.scope_distance == 1]
        )
        assert 1 == len(
            [
                e
                for e in events
                if e.type == "DNS_NAME"
                and e.data == "walmart.cn"
                and str(e.module) == "TARGET"
                and e.scope_distance == 0
            ]
        )
        assert 1 == len(
            [
                e
                for e in events
                if e.type == "DNS_NAME"
                and e.data == "asdf.walmart.cn"
                and str(e.module) == "TXT"
                and e.scope_distance == 0
                and "wildcard-possible" in e.tags
                and "a-wildcard-possible" in e.tags
            ]
        )
        assert 1 == len(
            [
                e
                for e in events
                if e.type == "RAW_DNS_RECORD"
                and e.data == {"host": "walmart.cn", "type": "TXT", "answer": '"asdf.walmart.cn"'}
            ]
        )



import ipaddress
from contextlib import suppress

from bbot.errors import ValidationError
from bbot.core.helpers.dns.engine import all_rdtypes
from bbot.core.helpers.dns.helpers import extract_targets
from bbot.modules.base import BaseInterceptModule, BaseModule


class DNSResolve(BaseInterceptModule):
    watched_events = ["*"]
    _priority = 1
    scope_distance_modifier = None

    class HostModule(BaseModule):
        _name = "host"
        _type = "internal"

    @property
    def module_threads(self):
        return self.dns_config.get("threads", 25)

    async def setup(self):
        self.dns_config = self.scan.config.get("dns", {})
        self.dns_disable = self.dns_config.get("disable", False)
        if self.dns_disable:
            return None, "DNS resolution is disabled in the config"

        self.minimal = self.dns_config.get("minimal", False)
        self.minimal_rdtypes = ("A", "AAAA", "CNAME")
        if self.minimal:
            self.non_minimal_rdtypes = ()
        else:
            self.non_minimal_rdtypes = tuple([t for t in all_rdtypes if t not in self.minimal_rdtypes])
        self.dns_search_distance = max(0, int(self.dns_config.get("search_distance", 1)))
        self._emit_raw_records = None

        self.host_module = self.HostModule(self.scan)
        self.children_emitted = set()
        self.children_emitted_raw = set()
        self.hosts_resolved = set()

        return True

    async def filter_event(self, event):
        if (not event.host) or (event.type in ("IP_RANGE",)):
            return False, "event does not have host attribute"
        return True

    async def handle_event(self, event, **kwargs):
        event_is_ip = self.helpers.is_ip(event.host)
        if event_is_ip:
            minimal_rdtypes = ("PTR",)
            non_minimal_rdtypes = ()
        else:
            minimal_rdtypes = self.minimal_rdtypes
            non_minimal_rdtypes = self.non_minimal_rdtypes

        # first, we find or create the main DNS_NAME or IP_ADDRESS associated with this event
        main_host_event, whitelisted, blacklisted, new_event = self.get_dns_parent(event)
        original_tags = set(event.tags)

        # minimal resolution - first, we resolve A/AAAA records for scope purposes
        if new_event or event is main_host_event:
            await self.resolve_event(main_host_event, types=minimal_rdtypes)
            # are any of its IPs whitelisted/blacklisted?
            whitelisted, blacklisted = self.check_scope(main_host_event)
            if whitelisted and event.scope_distance > 0:
                self.debug(f"Making {main_host_event} in-scope because it resolves to an in-scope resource (A/AAAA)")
                main_host_event.scope_distance = 0

        # abort if the event resolves to something blacklisted
        if blacklisted:
            return False, "it has a blacklisted DNS record"

        if not event_is_ip:
            # if the event is within our dns search distance, resolve the rest of our records
            if main_host_event.scope_distance < self._dns_search_distance:
                await self.resolve_event(main_host_event, types=non_minimal_rdtypes)
                # check for wildcards if the event is within the scan's search distance
                if new_event and main_host_event.scope_distance <= self.scan.scope_search_distance:
                    await self.handle_wildcard_event(main_host_event)

        # if there weren't any DNS children and it's not an IP address, tag as unresolved
        if not main_host_event.raw_dns_records and not event_is_ip:
            main_host_event.add_tag("unresolved")
            main_host_event.type = "DNS_NAME_UNRESOLVED"

        # main_host_event.add_tag(f"resolve-distance-{main_host_event.dns_resolve_distance}")

        dns_tags = main_host_event.tags.difference(original_tags)

        dns_resolve_distance = getattr(main_host_event, "dns_resolve_distance", 0)
        runaway_dns = dns_resolve_distance >= self.helpers.dns.runaway_limit
        if runaway_dns:
            # kill runaway DNS chains
            self.debug(
                f"Skipping DNS children for {event} because their DNS resolve distances would be greater than the configured value for this scan ({self.helpers.dns.runaway_limit})"
            )
            main_host_event.add_tag(f"runaway-dns-{dns_resolve_distance}")
        else:
            # emit dns children
            await self.emit_dns_children_raw(main_host_event, dns_tags)
            if not self.minimal:
                await self.emit_dns_children(main_host_event)

            # emit the main DNS_NAME or IP_ADDRESS
            if (
                new_event
                and event is not main_host_event
                and main_host_event.scope_distance <= self._dns_search_distance
            ):
                await self.emit_event(main_host_event)

        # transfer scope distance to event
        event.scope_distance = main_host_event.scope_distance
        event._resolved_hosts = main_host_event.resolved_hosts

    async def handle_wildcard_event(self, event):
        rdtypes = tuple(event.raw_dns_records)
        wildcard_rdtypes = await self.helpers.is_wildcard(
            event.host, rdtypes=rdtypes, raw_dns_records=event.raw_dns_records
        )
        for rdtype, (is_wildcard, wildcard_host) in wildcard_rdtypes.items():
            if is_wildcard == False:
                continue
            elif is_wildcard == True:
                event.add_tag("wildcard")
                wildcard_tag = "wildcard"
            else:
                event.add_tag(f"wildcard-{is_wildcard}")
                wildcard_tag = f"wildcard-{is_wildcard}"
            event.add_tag(f"{rdtype}-{wildcard_tag}")

        # wildcard event modification (www.evilcorp.com --> _wildcard.evilcorp.com)
        if wildcard_rdtypes and not "target" in event.tags:
            # these are the rdtypes that have wildcards
            wildcard_rdtypes_set = set(wildcard_rdtypes)
            # consider the event a full wildcard if all its records are wildcards
            event_is_wildcard = False
            if wildcard_rdtypes_set:
                event_is_wildcard = all(r[0] == True for r in wildcard_rdtypes.values())

            if event_is_wildcard:
                if event.type in ("DNS_NAME",) and not "_wildcard" in event.data.split("."):
                    wildcard_parent = self.helpers.parent_domain(event.host)
                    for rdtype, (_is_wildcard, _parent_domain) in wildcard_rdtypes.items():
                        if _is_wildcard:
                            wildcard_parent = _parent_domain
                            break
                    wildcard_data = f"_wildcard.{wildcard_parent}"
                    if wildcard_data != event.data:
                        self.debug(f'Wildcard detected, changing event.data "{event.data}" --> "{wildcard_data}"')
                        event.data = wildcard_data

    async def emit_dns_children(self, event):
        for rdtype, children in event.dns_children.items():
            module = self._make_dummy_module(rdtype)
            for child_host in children:
                try:
                    child_event = self.scan.make_event(
                        child_host,
                        "DNS_NAME",
                        module=module,
                        parent=event,
                        context=f"{rdtype} record for {event.host} contains {{event.type}}: {{event.host}}",
                    )
                except ValidationError as e:
                    self.warning(f'Event validation failed for DNS child of {event}: "{child_host}" ({rdtype}): {e}')
                    continue

                child_hash = hash(f"{event.host}:{module}:{child_host}")
                # if we haven't emitted this one before
                if child_hash not in self.children_emitted:
                    # and it's either in-scope or inside our dns search distance
                    if self.preset.in_scope(child_host) or child_event.scope_distance <= self._dns_search_distance:
                        self.children_emitted.add(child_hash)
                        # if it's a hostname and it's only one hop away, mark it as affiliate
                        if child_event.type == "DNS_NAME" and child_event.scope_distance == 1:
                            child_event.add_tag("affiliate")
                        self.debug(f"Queueing DNS child for {event}: {child_event}")
                        await self.emit_event(child_event)

    async def emit_dns_children_raw(self, event, dns_tags):
        for rdtype, answers in event.raw_dns_records.items():
            rdtype_lower = rdtype.lower()
            tags = {t for t in dns_tags if rdtype_lower in t.split("-")}
            if self.emit_raw_records and rdtype not in ("A", "AAAA", "CNAME", "PTR"):
                for answer in answers:
                    text_answer = answer.to_text()
                    child_hash = hash(f"{event.host}:{rdtype}:{text_answer}")
                    if child_hash not in self.children_emitted_raw:
                        self.children_emitted_raw.add(child_hash)
                        await self.emit_event(
                            {"host": str(event.host), "type": rdtype, "answer": text_answer},
                            "RAW_DNS_RECORD",
                            parent=event,
                            tags=tags,
                            context=f"{rdtype} lookup on {{event.parent.host}} produced {{event.type}}",
                        )

    def check_scope(self, event):
        whitelisted = False
        blacklisted = False
        dns_children = getattr(event, "dns_children", {})
        for rdtype in ("A", "AAAA", "CNAME"):
            hosts = dns_children.get(rdtype, [])
            # update resolved hosts
            event.resolved_hosts.update(hosts)
            for host in hosts:
                # having a CNAME to an in-scope host doesn't make you in-scope
                if rdtype != "CNAME":
                    if not whitelisted:
                        with suppress(ValidationError):
                            if self.scan.whitelisted(host):
                                whitelisted = True
                                event.add_tag(f"dns-whitelisted-{rdtype}")
                # but a CNAME to a blacklisted host means you're blacklisted
                if not blacklisted:
                    with suppress(ValidationError):
                        if self.scan.blacklisted(host):
                            blacklisted = True
                            event.add_tag("blacklisted")
                            event.add_tag(f"dns-blacklisted-{rdtype}")
        if blacklisted:
            whitelisted = False
        return whitelisted, blacklisted

    async def resolve_event(self, event, types):
        if not types:
            return
        event_host = str(event.host)
        queries = [(event_host, rdtype) for rdtype in types]
        dns_errors = {}
        async for (query, rdtype), (answers, errors) in self.helpers.dns.resolve_raw_batch(queries):
            # errors
            try:
                dns_errors[rdtype].update(errors)
            except KeyError:
                dns_errors[rdtype] = set(errors)
            for answer in answers:
                event.add_tag(f"{rdtype}-record")
                # raw dnspython answers
                try:
                    event.raw_dns_records[rdtype].add(answer)
                except KeyError:
                    event.raw_dns_records[rdtype] = {answer}
                # hosts
                for _rdtype, host in extract_targets(answer):
                    try:
                        event.dns_children[_rdtype].add(host)
                    except KeyError:
                        event.dns_children[_rdtype] = {host}
                    # check for private IPs
                    try:
                        ip = ipaddress.ip_address(host)
                        if ip.is_private:
                            event.add_tag("private-ip")
                    except ValueError:
                        continue

        # tag event with errors
        for rdtype, errors in dns_errors.items():
            # only consider it an error if there weren't any results for that rdtype
            if errors and not rdtype in event.dns_children:
                event.add_tag(f"{rdtype}-error")

    def get_dns_parent(self, event):
        """
        Get the first parent DNS_NAME / IP_ADDRESS of an event. If one isn't found, create it.
        """
        for parent in event.get_parents(include_self=True):
            if parent.host == event.host and parent.type in ("IP_ADDRESS", "DNS_NAME", "DNS_NAME_UNRESOLVED"):
                blacklisted = any(t.startswith("dns-blacklisted-") for t in parent.tags)
                whitelisted = any(t.startswith("dns-whitelisted-") for t in parent.tags)
                new_event = parent is event
                return parent, whitelisted, blacklisted, new_event
        tags = set()
        if "target" in event.tags:
            tags.add("target")
        return (
            self.scan.make_event(
                event.host,
                "DNS_NAME",
                module=self.host_module,
                parent=event,
                context="{event.parent.type} has host {event.type}: {event.host}",
                tags=tags,
            ),
            None,
            None,
            True,
        )

    @property
    def emit_raw_records(self):
        if self._emit_raw_records is None:
            watching_raw_records = any(
                ["RAW_DNS_RECORD" in m.get_watched_events() for m in self.scan.modules.values()]
            )
            omitted_event_types = self.scan.config.get("omit_event_types", [])
            omit_raw_records = "RAW_DNS_RECORD" in omitted_event_types
            self._emit_raw_records = watching_raw_records or not omit_raw_records
        return self._emit_raw_records

    @property
    def _dns_search_distance(self):
        return max(self.scan.scope_search_distance, self.dns_search_distance)

    def _make_dummy_module(self, name):
        try:
            dummy_module = self.scan.dummy_modules[name]
        except KeyError:
            dummy_module = self.scan._make_dummy_module(name=name, _type="DNS")
            dummy_module._priority = 4
            dummy_module.suppress_dupes = False
            self.scan.dummy_modules[name] = dummy_module
        return dummy_module



from .base import ModuleTestBase


class TestEmailFormat(ModuleTestBase):
    async def setup_before_prep(self, module_test):
        module_test.httpx_mock.add_response(
            url="https://www.email-format.com/d/blacklanternsecurity.com/",
            text="<p>info@blacklanternsecurity.com</a>",
        )

    def check(self, module_test, events):
        assert any(e.data == "info@blacklanternsecurity.com" for e in events), "Failed to detect email"



import logging
import warnings
from pathlib import Path
from bs4 import BeautifulSoup

from bbot.core.engine import EngineClient
from bbot.core.helpers.misc import truncate_filename
from bbot.errors import WordlistError, CurlError, WebError

from bs4 import MarkupResemblesLocatorWarning
from bs4.builder import XMLParsedAsHTMLWarning

from .engine import HTTPEngine

warnings.filterwarnings("ignore", category=XMLParsedAsHTMLWarning)
warnings.filterwarnings("ignore", category=MarkupResemblesLocatorWarning)

log = logging.getLogger("bbot.core.helpers.web")


class WebHelper(EngineClient):

    SERVER_CLASS = HTTPEngine
    ERROR_CLASS = WebError

    """
    Main utility class for managing HTTP operations in BBOT. It serves as a wrapper around the BBOTAsyncClient,
    which itself is a subclass of httpx.AsyncClient. The class provides functionalities to make HTTP requests,
    download files, and handle cached wordlists.

    Attributes:
        parent_helper (object): The parent helper object containing scan configurations.
        http_debug (bool): Flag to indicate whether HTTP debugging is enabled.
        ssl_verify (bool): Flag to indicate whether SSL verification is enabled.
        web_client (BBOTAsyncClient): An instance of BBOTAsyncClient for making HTTP requests.
        client_only_options (tuple): A tuple of options only applicable to the web client.

    Examples:
        Basic web request:
        >>> response = await self.helpers.request("https://www.evilcorp.com")

        Download file:
        >>> filename = await self.helpers.download("https://www.evilcorp.com/passwords.docx")

        Download wordlist (cached for 30 days by default):
        >>> filename = await self.helpers.wordlist("https://www.evilcorp.com/wordlist.txt")
    """

    def __init__(self, parent_helper):
        self.parent_helper = parent_helper
        self.preset = self.parent_helper.preset
        self.config = self.preset.config
        self.web_config = self.config.get("web", {})
        self.web_spider_depth = self.web_config.get("spider_depth", 1)
        self.web_spider_distance = self.web_config.get("spider_distance", 0)
        self.web_clients = {}
        self.target = self.preset.target
        self.ssl_verify = self.config.get("ssl_verify", False)
        engine_debug = self.config.get("engine", {}).get("debug", False)
        super().__init__(
            server_kwargs={"config": self.config, "target": self.parent_helper.preset.target.radix_only},
            debug=engine_debug,
        )

    def AsyncClient(self, *args, **kwargs):
        # cache by retries to prevent unwanted accumulation of clients
        # (they are not garbage-collected)
        retries = kwargs.get("retries", 1)
        try:
            return self.web_clients[retries]
        except KeyError:
            from .client import BBOTAsyncClient

            client = BBOTAsyncClient.from_config(self.config, self.target, *args, persist_cookies=False, **kwargs)
            self.web_clients[client.retries] = client
            return client

    async def request(self, *args, **kwargs):
        """
        Asynchronous function for making HTTP requests, intended to be the most basic web request function
        used widely across BBOT and within this helper class. Handles various exceptions and timeouts
        that might occur during the request.

        This function automatically respects the scan's global timeout, proxy, headers, etc.
        Headers you specify will be merged with the scan's. Your arguments take ultimate precedence,
        meaning you can override the scan's values if you want.

        Args:
            url (str): The URL to send the request to.
            method (str, optional): The HTTP method to use for the request. Defaults to 'GET'.
            headers (dict, optional): Dictionary of HTTP headers to send with the request.
            params (dict, optional): Dictionary, list of tuples, or bytes to send in the query string.
            cookies (dict, optional): Dictionary or CookieJar object containing cookies.
            json (Any, optional): A JSON serializable Python object to send in the body.
            data (dict, optional): Dictionary, list of tuples, or bytes to send in the body.
            files (dict, optional): Dictionary of 'name': file-like-objects for multipart encoding upload.
            auth (tuple, optional): Auth tuple to enable Basic/Digest/Custom HTTP auth.
            timeout (float, optional): The maximum time to wait for the request to complete.
            proxies (dict, optional): Dictionary mapping protocol schemes to proxy URLs.
            allow_redirects (bool, optional): Enables or disables redirection. Defaults to None.
            stream (bool, optional): Enables or disables response streaming.
            raise_error (bool, optional): Whether to raise exceptions for HTTP connect, timeout errors. Defaults to False.
            client (httpx.AsyncClient, optional): A specific httpx.AsyncClient to use for the request. Defaults to self.web_client.
            cache_for (int, optional): Time in seconds to cache the request. Not used currently. Defaults to None.

        Raises:
            httpx.TimeoutException: If the request times out.
            httpx.ConnectError: If the connection fails.
            httpx.RequestError: For other request-related errors.

        Returns:
            httpx.Response or None: The HTTP response object returned by the httpx library.

        Examples:
            >>> response = await self.helpers.request("https://www.evilcorp.com")

            >>> response = await self.helpers.request("https://api.evilcorp.com/", method="POST", data="stuff")

        Note:
            If the web request fails, it will return None unless `raise_error` is `True`.
        """
        raise_error = kwargs.get("raise_error", False)
        result = await self.run_and_return("request", *args, **kwargs)
        if isinstance(result, dict) and "_request_error" in result:
            if raise_error:
                error_msg = result["_request_error"]
                response = result["_response"]
                error = self.ERROR_CLASS(error_msg)
                error.response = response
                raise error
        return result

    async def request_batch(self, urls, *args, **kwargs):
        """
        Given a list of URLs, request them in parallel and yield responses as they come in.

        Args:
            urls (list[str]): List of URLs to visit
            *args: Positional arguments to pass through to httpx
            **kwargs: Keyword arguments to pass through to httpx

        Examples:
            >>> async for url, response in self.helpers.request_batch(urls, headers={"X-Test": "Test"}):
            >>>     if response is not None and response.status_code == 200:
            >>>         self.hugesuccess(response)
        """
        agen = self.run_and_yield("request_batch", urls, *args, **kwargs)
        while 1:
            try:
                yield await agen.__anext__()
            except (StopAsyncIteration, GeneratorExit):
                await agen.aclose()
                break

    async def request_custom_batch(self, urls_and_kwargs):
        """
        Make web requests in parallel with custom options for each request. Yield responses as they come in.

        Similar to `request_batch` except it allows individual arguments for each URL.

        Args:
            urls_and_kwargs (list[tuple]): List of tuples in the format: (url, kwargs, custom_tracker)
                where custom_tracker is an optional value for your own internal use. You may use it to
                help correlate requests, etc.

        Examples:
            >>> urls_and_kwargs = [
            >>>     ("http://evilcorp.com/1", {"method": "GET"}, "request-1"),
            >>>     ("http://evilcorp.com/2", {"method": "POST"}, "request-2"),
            >>> ]
            >>> async for url, kwargs, custom_tracker, response in self.helpers.request_custom_batch(
            >>>     urls_and_kwargs
            >>> ):
            >>>     if response is not None and response.status_code == 200:
            >>>         self.hugesuccess(response)
        """
        agen = self.run_and_yield("request_custom_batch", urls_and_kwargs)
        while 1:
            try:
                yield await agen.__anext__()
            except (StopAsyncIteration, GeneratorExit):
                await agen.aclose()
                break

    async def download(self, url, **kwargs):
        """
        Asynchronous function for downloading files from a given URL. Supports caching with an optional
        time period in hours via the "cache_hrs" keyword argument. In case of successful download,
        returns the full path of the saved filename. If the download fails, returns None.

        Args:
            url (str): The URL of the file to download.
            filename (str, optional): The filename to save the downloaded file as.
                If not provided, will generate based on URL.
            max_size (str or int): Maximum filesize as a string ("5MB") or integer in bytes.
            cache_hrs (float, optional): The number of hours to cache the downloaded file.
                A negative value disables caching. Defaults to -1.
            method (str, optional): The HTTP method to use for the request, defaults to 'GET'.
            raise_error (bool, optional): Whether to raise exceptions for HTTP connect, timeout errors. Defaults to False.
            **kwargs: Additional keyword arguments to pass to the httpx request.

        Returns:
            Path or None: The full path of the downloaded file as a Path object if successful, otherwise None.

        Examples:
            >>> filepath = await self.helpers.download("https://www.evilcorp.com/passwords.docx", cache_hrs=24)
        """
        success = False
        raise_error = kwargs.get("raise_error", False)
        filename = kwargs.pop("filename", self.parent_helper.cache_filename(url))
        filename = truncate_filename(Path(filename).resolve())
        kwargs["filename"] = filename
        max_size = kwargs.pop("max_size", None)
        if max_size is not None:
            max_size = self.parent_helper.human_to_bytes(max_size)
            kwargs["max_size"] = max_size
        cache_hrs = float(kwargs.pop("cache_hrs", -1))
        if cache_hrs > 0 and self.parent_helper.is_cached(url):
            log.debug(f"{url} is cached at {self.parent_helper.cache_filename(url)}")
            success = True
        else:
            result = await self.run_and_return("download", url, **kwargs)
            if isinstance(result, dict) and "_download_error" in result:
                if raise_error:
                    error_msg = result["_download_error"]
                    response = result["_response"]
                    error = self.ERROR_CLASS(error_msg)
                    error.response = response
                    raise error
            elif result:
                success = True

        if success:
            return filename

    async def wordlist(self, path, lines=None, **kwargs):
        """
        Asynchronous function for retrieving wordlists, either from a local path or a URL.
        Allows for optional line-based truncation and caching. Returns the full path of the wordlist
        file or a truncated version of it.

        Args:
            path (str): The local or remote path of the wordlist.
            lines (int, optional): Number of lines to read from the wordlist.
                If specified, will return a truncated wordlist with this many lines.
            cache_hrs (float, optional): Number of hours to cache the downloaded wordlist.
                Defaults to 720 hours (30 days) for remote wordlists.
            **kwargs: Additional keyword arguments to pass to the 'download' function for remote wordlists.

        Returns:
            Path: The full path of the wordlist (or its truncated version) as a Path object.

        Raises:
            WordlistError: If the path is invalid or the wordlist could not be retrieved or found.

        Examples:
            Fetching full wordlist
            >>> wordlist_path = await self.helpers.wordlist("https://www.evilcorp.com/wordlist.txt")

            Fetching and truncating to the first 100 lines
            >>> wordlist_path = await self.helpers.wordlist("/root/rockyou.txt", lines=100)
        """
        if not path:
            raise WordlistError(f"Invalid wordlist: {path}")
        if not "cache_hrs" in kwargs:
            kwargs["cache_hrs"] = 720
        if self.parent_helper.is_url(path):
            filename = await self.download(str(path), **kwargs)
            if filename is None:
                raise WordlistError(f"Unable to retrieve wordlist from {path}")
        else:
            filename = Path(path).resolve()
            if not filename.is_file():
                raise WordlistError(f"Unable to find wordlist at {path}")

        if lines is None:
            return filename
        else:
            lines = int(lines)
            with open(filename) as f:
                read_lines = f.readlines()
            cache_key = f"{filename}:{lines}"
            truncated_filename = self.parent_helper.cache_filename(cache_key)
            with open(truncated_filename, "w") as f:
                for line in read_lines[:lines]:
                    f.write(line)
            return truncated_filename

    async def curl(self, *args, **kwargs):
        """
        An asynchronous function that runs a cURL command with specified arguments and options.

        This function constructs and executes a cURL command based on the provided parameters.
        It offers support for various cURL options such as headers, post data, and cookies.

        Args:
            *args: Variable length argument list for positional arguments. Unused in this function.
            url (str): The URL for the cURL request. Mandatory.
            raw_path (bool, optional): If True, activates '--path-as-is' in cURL. Defaults to False.
            headers (dict, optional): A dictionary of HTTP headers to include in the request.
            ignore_bbot_global_settings (bool, optional): If True, ignores the global settings of BBOT. Defaults to False.
            post_data (dict, optional): A dictionary containing data to be sent in the request body.
            method (str, optional): The HTTP method to use for the request (e.g., 'GET', 'POST').
            cookies (dict, optional): A dictionary of cookies to include in the request.
            path_override (str, optional): Overrides the request-target to use in the HTTP request line.
            head_mode (bool, optional): If True, includes '-I' to fetch headers only. Defaults to None.
            raw_body (str, optional): Raw string to be sent in the body of the request.
            **kwargs: Arbitrary keyword arguments that will be forwarded to the HTTP request function.

        Returns:
            str: The output of the cURL command.

        Raises:
            CurlError: If 'url' is not supplied.

        Examples:
            >>> output = await curl(url="https://example.com", headers={"X-Header": "Wat"})
            >>> print(output)
        """
        url = kwargs.get("url", "")

        if not url:
            raise CurlError("No URL supplied to CURL helper")

        curl_command = ["curl", url, "-s"]

        raw_path = kwargs.get("raw_path", False)
        if raw_path:
            curl_command.append("--path-as-is")

        # respect global ssl verify settings
        if self.ssl_verify is not True:
            curl_command.append("-k")

        headers = kwargs.get("headers", {})

        ignore_bbot_global_settings = kwargs.get("ignore_bbot_global_settings", False)

        if ignore_bbot_global_settings:
            log.debug("ignore_bbot_global_settings enabled. Global settings will not be applied")
        else:
            http_timeout = self.parent_helper.web_config.get("http_timeout", 20)
            user_agent = self.parent_helper.web_config.get("user_agent", "BBOT")

            if "User-Agent" not in headers:
                headers["User-Agent"] = user_agent

            # only add custom headers if the URL is in-scope
            if self.parent_helper.preset.in_scope(url):
                for hk, hv in self.web_config.get("http_headers", {}).items():
                    headers[hk] = hv

            # add the timeout
            if not "timeout" in kwargs:
                timeout = http_timeout

            curl_command.append("-m")
            curl_command.append(str(timeout))

        for k, v in headers.items():
            if isinstance(v, list):
                for x in v:
                    curl_command.append("-H")
                    curl_command.append(f"{k}: {x}")

            else:
                curl_command.append("-H")
                curl_command.append(f"{k}: {v}")

        post_data = kwargs.get("post_data", {})
        if len(post_data.items()) > 0:
            curl_command.append("-d")
            post_data_str = ""
            for k, v in post_data.items():
                post_data_str += f"&{k}={v}"
            curl_command.append(post_data_str.lstrip("&"))

        method = kwargs.get("method", "")
        if method:
            curl_command.append("-X")
            curl_command.append(method)

        cookies = kwargs.get("cookies", "")
        if cookies:
            curl_command.append("-b")
            cookies_str = ""
            for k, v in cookies.items():
                cookies_str += f"{k}={v}; "
            curl_command.append(f'{cookies_str.rstrip(" ")}')

        path_override = kwargs.get("path_override", None)
        if path_override:
            curl_command.append("--request-target")
            curl_command.append(f"{path_override}")

        head_mode = kwargs.get("head_mode", None)
        if head_mode:
            curl_command.append("-I")

        raw_body = kwargs.get("raw_body", None)
        if raw_body:
            curl_command.append("-d")
            curl_command.append(raw_body)

        output = (await self.parent_helper.run(curl_command)).stdout
        return output

    def beautifulsoup(
        self,
        markup,
        features="html.parser",
        builder=None,
        parse_only=None,
        from_encoding=None,
        exclude_encodings=None,
        element_classes=None,
        **kwargs,
    ):
        """
        Naviate, Search, Modify, Parse, or PrettyPrint HTML Content.
        More information at https://beautiful-soup-4.readthedocs.io/en/latest/

        Args:
            markup: A string or a file-like object representing markup to be parsed.
            features: Desirable features of the parser to be used.
                This may be the name of a specific parser ("lxml",
                "lxml-xml", "html.parser", or "html5lib") or it may be
                the type of markup to be used ("html", "html5", "xml").
                Defaults to 'html.parser'.
            builder: A TreeBuilder subclass to instantiate (or instance to use)
                instead of looking one up based on `features`.
            parse_only: A SoupStrainer. Only parts of the document
                matching the SoupStrainer will be considered.
            from_encoding: A string indicating the encoding of the
                document to be parsed.
            exclude_encodings = A list of strings indicating
                encodings known to be wrong.
            element_classes = A dictionary mapping BeautifulSoup
                classes like Tag and NavigableString, to other classes you'd
                like to be instantiated instead as the parse tree is
                built.
            **kwargs = For backwards compatibility purposes.

        Returns:
            soup: An instance of the BeautifulSoup class

        Todo:
            - Write tests for this function

        Examples:
            >>> soup = self.helpers.beautifulsoup(event.data["body"], "html.parser")
            Perform an html parse of the 'markup' argument and return a soup instance

            >>> email_type = soup.find(type="email")
            Searches the soup instance for all occurances of the passed in argument
        """
        try:
            soup = BeautifulSoup(
                markup, features, builder, parse_only, from_encoding, exclude_encodings, element_classes, **kwargs
            )
            return soup
        except Exception as e:
            log.debug(f"Error parsing beautifulsoup: {e}")
            return False

    def response_to_json(self, response):
        """
        Convert web response to JSON object, similar to the output of `httpx -irr -json`
        """

        if response is None:
            return

        import mmh3
        from datetime import datetime
        from hashlib import md5, sha256
        from bbot.core.helpers.misc import tagify, urlparse, split_host_port, smart_decode

        request = response.request
        url = str(request.url)
        parsed_url = urlparse(url)
        netloc = parsed_url.netloc
        scheme = parsed_url.scheme.lower()
        host, port = split_host_port(f"{scheme}://{netloc}")

        raw_headers = "\r\n".join([f"{k}: {v}" for k, v in response.headers.items()])
        raw_headers_encoded = raw_headers.encode()

        headers = {}
        for k, v in response.headers.items():
            k = tagify(k, delimiter="_")
            headers[k] = v

        j = {
            "timestamp": datetime.now().isoformat(),
            "hash": {
                "body_md5": md5(response.content).hexdigest(),
                "body_mmh3": mmh3.hash(response.content),
                "body_sha256": sha256(response.content).hexdigest(),
                # "body_simhash": "TODO",
                "header_md5": md5(raw_headers_encoded).hexdigest(),
                "header_mmh3": mmh3.hash(raw_headers_encoded),
                "header_sha256": sha256(raw_headers_encoded).hexdigest(),
                # "header_simhash": "TODO",
            },
            "header": headers,
            "body": smart_decode(response.content),
            "content_type": headers.get("content_type", "").split(";")[0].strip(),
            "url": url,
            "host": str(host),
            "port": port,
            "scheme": scheme,
            "method": response.request.method,
            "path": parsed_url.path,
            "raw_header": raw_headers,
            "status_code": response.status_code,
        }

        return j



from bbot.modules.templates.subdomain_enum import subdomain_enum


class sitedossier(subdomain_enum):
    flags = ["subdomain-enum", "passive", "safe"]
    watched_events = ["DNS_NAME"]
    produced_events = ["DNS_NAME"]
    meta = {
        "description": "Query sitedossier.com for subdomains",
        "created_date": "2023-08-04",
        "author": "@TheTechromancer",
    }

    base_url = "http://www.sitedossier.com/parentdomain"
    max_pages = 10

    async def handle_event(self, event):
        query = self.make_query(event)
        async for hostname in self.query(query):
            try:
                hostname = self.helpers.validators.validate_host(hostname)
            except ValueError as e:
                self.verbose(e)
                continue
            if hostname and hostname.endswith(f".{query}") and not hostname == event.data:
                await self.emit_event(
                    hostname,
                    "DNS_NAME",
                    event,
                    abort_if=self.abort_if,
                    context=f'{{module}} searched sitedossier.com for "{query}" and found {{event.type}}: {{event.data}}',
                )

    async def query(self, query, parse_fn=None, request_fn=None):
        results = set()
        base_url = f"{self.base_url}/{self.helpers.quote(query)}"
        url = str(base_url)
        for i, page in enumerate(range(1, 100 * self.max_pages + 2, 100)):
            self.verbose(f"Fetching page #{i+1} for {query}")
            if page > 1:
                url = f"{base_url}/{page}"
            response = await self.helpers.request(url)
            if response is None:
                self.info(f'Query "{query}" failed (no response)')
                break
            if response.status_code == 302:
                self.verbose("Hit rate limit captcha")
                break
            for match in await self.helpers.re.finditer_multi(self.scan.dns_regexes, response.text):
                hostname = match.group().lower()
                if hostname and hostname not in results:
                    results.add(hostname)
                    yield hostname
            if '<a href="/parentdomain/' not in response.text:
                self.debug(f"Next page not found")
                break
